import base64
from datetime import datetime, timezone
import json
from typing_extensions import Annotated
from fastapi import Depends, HTTPException, Header, Security
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer
import httpx
import jwt
from syftbox.server.settings import ServerSettings, get_server_settings

bearer_scheme = HTTPBearer()

ACCESS_TOKEN = "access_token"
EMAIL_TOKEN = "email_token"


def _validate_jwt(server_settings: ServerSettings, token: str) -> dict:
    try:
        return jwt.decode(
            token,
            server_settings.jwt_secret.get_secret_value(),
            algorithms=[server_settings.jwt_algorithm],
        )
    except jwt.ExpiredSignatureError:
        raise HTTPException(
            status_code=401,
            detail="Token has expired",
            headers={"WWW-Authenticate": "Bearer"},
        )
    except Exception as e:
        raise HTTPException(
            status_code=401,
            detail="Invalid token",
            headers={"WWW-Authenticate": "Bearer"},
        )

def _generate_jwt(server_settings: ServerSettings, data: dict) -> str:
    return jwt.encode(
        data,
        server_settings.jwt_secret.get_secret_value(),
        algorithm=server_settings.jwt_algorithm,
    )


def generate_access_token(server_settings: ServerSettings, email: str) -> str:
    data = {
        "email": email,
        "type": ACCESS_TOKEN,
        "iat": datetime.now(tz=timezone.utc),
    }
    if server_settings.jwt_access_token_exp:
        data["exp"] = data["iat"] + server_settings.jwt_access_token_exp
    return _generate_jwt(server_settings, data)


def generate_email_token(server_settings: ServerSettings, email: str) -> str:
    data = {
        "email": email,
        "type": EMAIL_TOKEN,
        "iat": datetime.now(tz=timezone.utc),
    }
    if server_settings.jwt_email_token_exp:
        data["exp"] = data["iat"] + server_settings.jwt_email_token_exp
    return _generate_jwt(server_settings, data)


def validate_access_token(server_settings: ServerSettings, token: str) -> dict:
    data = _validate_jwt(server_settings, token)
    if data["type"] != ACCESS_TOKEN:
        raise HTTPException(
            status_code=401,
            detail="Invalid token type",
            headers={"WWW-Authenticate": "Bearer"},
        )
    return data


def validate_email_token(server_settings: ServerSettings, token: str) -> dict:
    data = _validate_jwt(server_settings, token)
    if data["type"] != EMAIL_TOKEN:
        raise HTTPException(
            status_code=401,
            detail="Invalid token type",
            headers={"WWW-Authenticate": "Bearer"},
        )
    return data


def get_user_from_email_token(
    credentials: Annotated[HTTPAuthorizationCredentials, Security(bearer_scheme)],
    server_settings: Annotated[ServerSettings, Depends(get_server_settings)],
) -> str:
    payload = validate_email_token(server_settings, credentials.credentials)
    return payload["email"]


def get_current_user(
    credentials: Annotated[HTTPAuthorizationCredentials, Security(bearer_scheme)],
    server_settings: Annotated[ServerSettings, Depends(get_server_settings)],
) -> str:
    payload = validate_access_token(server_settings, credentials.credentials)
    return payload["email"]


import uuid
from pathlib import Path

from locust import FastHttpUser, between, task

import syftbox.client.exceptions
from syftbox.client.plugins.sync import consumer, endpoints
from syftbox.server.sync.hash import hash_file

file_name = Path("loadtest.txt")


class SyftBoxUser(FastHttpUser):
    network_timeout = 5.0
    connection_timeout = 5.0
    wait_time = between(0.5, 1.5)

    def on_start(self):
        self.datasites = []
        self.email = "aziz@openmined.org"
        self.remote_state: dict[str, list[endpoints.FileMetadata]] = {}

        # patch client for update_remote function
        self.client.sync_folder = Path(".")
        self.client.server_client = self.client

        self.filepath = self.init_file()

    def init_file(self) -> Path:
        # create a file on local and send to server
        filepath = self.client.sync_folder / file_name
        filepath.parent.mkdir(parents=True, exist_ok=True)
        filepath.touch()
        filepath.write_text(uuid.uuid4().hex)
        local_syncstate = hash_file(filepath.absolute(), root_dir=filepath.parent.absolute())
        try:
            endpoints.create(self.client, local_syncstate.path, filepath.read_bytes())
        except syftbox.client.exceptions.SyftServerError:
            pass
        return filepath

    @task
    def sync_datasites(self):
        remote_datasite_states = endpoints.get_datasite_states(
            self.client,
            email=self.email,
        )
        # logger.info(f"Syncing {len(remote_datasite_states)} datasites")
        all_files = []
        for email, remote_state in remote_datasite_states.items():
            all_files.extend(remote_state)

        all_paths = [str(f.path) for f in all_files][:10]
        endpoints.download_bulk(
            self.client,
            all_paths,
        )

    @task
    def apply_diff(self):
        self.filepath.write_text(uuid.uuid4().hex)
        local_syncstate = hash_file(self.filepath, root_dir=self.client.sync_folder)
        remote_syncstate = endpoints.get_metadata(self.client, local_syncstate.path)

        consumer.update_remote(
            self.client,
            local_syncstate=local_syncstate,
            remote_syncstate=remote_syncstate,
        )

    @task
    def download(self):
        endpoints.download(self.client, self.filepath)



import re
import shutil
import tempfile

from typing_extensions import Tuple

from syftbox.lib.types import PathLike, to_path

DIR_NOT_EMPTY = "Directory is not empty"


def is_valid_dir(path: PathLike, check_empty=True, check_writable=True) -> Tuple[bool, str]:
    try:
        if not path:
            return False, "Empty path"

        # Convert to Path object if string
        dir_path = to_path(path)

        # Must not be a reserved path
        if dir_path.is_reserved():
            return False, "Reserved path"

        if dir_path.exists():
            if not dir_path.is_dir():
                return False, "Path is not a directory"

            if check_empty and any(dir_path.iterdir()):
                return False, DIR_NOT_EMPTY
        elif check_writable:
            # Try to create a temporary file to test write permissions on parent
            try:
                dir_path.mkdir(parents=True, exist_ok=True)
                testfile = tempfile.TemporaryFile(dir=dir_path)
                testfile.close()
                shutil.rmtree(dir_path)
            except Exception as e:
                return False, str(e)

        # all checks passed
        return True, ""
    except Exception as e:
        return False, str(e)


def is_valid_email(email: str) -> bool:
    # Define a regex pattern for a valid email
    # from: https://stackoverflow.com/a/21608610
    email_regex = r"\w+([-+.']\w+)*@\w+([-.]\w+)*\.\w+([-.]\w+)*"

    # Use the match method to check if the email fits the pattern
    if re.match(email_regex, email):
        return True
    return False



from typing import Optional
from fastapi import APIRouter, Depends, HTTPException
from pydantic import BaseModel, EmailStr

from syftbox.lib.email import send_token_email
from syftbox.server.analytics import log_analytics_event
from syftbox.server.settings import ServerSettings, get_server_settings
from syftbox.server.users.auth import generate_access_token, generate_email_token, get_user_from_email_token, get_current_user

router = APIRouter(prefix="/auth", tags=["authentication"])


class EmailTokenRequest(BaseModel):
    email: EmailStr

class EmailTokenResponse(BaseModel):
    email_token: Optional[str] = None

class AccessTokenResponse(BaseModel):
    access_token: str


@router.post("/request_email_token")
def get_token(req: EmailTokenRequest, server_settings: ServerSettings = Depends(get_server_settings)) -> EmailTokenResponse:
    """
    Send an email token to the user's email address

    if auth is disabled, the token will be returned in the response as a base64 encoded json string
    """
    email = req.email
    token = generate_email_token(server_settings, email)

    response = EmailTokenResponse()
    if server_settings.auth_enabled:
        send_token_email(server_settings, email, token)
    else:
        # Only return token if auth is disabled, it will be a base64 encoded json string
        response.email_token = token

    return response


@router.post("/validate_email_token")
def validate_email_token(
    email: str,
    email_from_token: str = Depends(get_user_from_email_token),
    server_settings: ServerSettings = Depends(get_server_settings),
) -> AccessTokenResponse:
    """
    Validate the email token and return a matching access token

    Args:
        email (str, optional): The user email, extracted from the email token. Defaults to Depends(get_user_from_email_token).
        server_settings (ServerSettings, optional): server settings. Defaults to Depends(get_server_settings).

    Returns:
        AccessTokenResponse: access token
    """
    if email_from_token != email:
        raise HTTPException(status_code=401, detail="This email token is not for this email address")

    access_token = generate_access_token(server_settings, email)
    return AccessTokenResponse(access_token=access_token)

class WhoAmIResponse(BaseModel):
    email: str

@router.post("/whoami")
def whoami(
    email: str = Depends(get_current_user),
) -> WhoAmIResponse:
    """
    Get the current users email.
    If the token is not valid or outdated, get_current_user will raise 401 Unauthorized.

    Returns:
        str: email
    """
    log_analytics_event("/auth/whoami", email=email)
    return WhoAmIResponse(email=email)


import asyncio
import os
import shutil
import sys
from asyncio.subprocess import Process
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List

import httpx
import pytest_asyncio
from loguru import logger

logger.remove()
logger.add(sys.stderr, format="<level>{level: <8}</level> | <level>{message}</level>", colorize=True)
logger.level("SUCCESS", color="<b><green>")
logger.level("INFO", color="<b><cyan>")
logger.level("DEBUG", color="<blue>")


class E2ETestError(Exception):
    pass


class E2ETimeoutError(E2ETestError):
    pass


@dataclass
class Server:
    port: int = 5001
    env: Dict[str, str] = field(default_factory=dict)


@dataclass
class Client:
    name: str
    port: int
    server_port: int = 5001
    data_dir: Path = field(default_factory=Path.cwd)  # Set by E2EContext
    env: Dict[str, str] = field(default_factory=dict)
    apps: List[str] = field(default_factory=list)

    @property
    def email(self):
        return f"{self.name}@openmined.org"

    @property
    def datasite_dir(self):
        """data_dir/datasites"""
        return self.data_dir / "datasites"

    @property
    def api_dir(self):
        """data_dir/apis"""
        return self.data_dir / "apis"

    @property
    def private_dir(self):
        """data_dir/private"""
        return self.data_dir / "private"

    @property
    def my_datasite(self):
        """data_dir/datasites/{email}"""
        return self.datasite_dir / self.email

    @property
    def public_dir(self):
        """data_dir/datasites/{email}/public"""
        return self.my_datasite / "public"

    def api_path(self, api_name: str):
        """data_dir/apis/{api_name}"""
        return self.api_dir / api_name

    def api_data_dir(self, app_name: str):
        """data_dir/datasites/{email}/api_data/{app_name}"""
        return self.my_datasite / "api_data" / app_name


class E2EContext:
    def __init__(self, e2e_name: str, server: Server, clients: List[Client]):
        self.e2e_name = e2e_name
        self.test_dir = Path.cwd() / ".e2e" / e2e_name
        self.server = server
        self.clients = clients
        self.__procs: List[Process] = []

    def reset_test_dir(self):
        shutil.rmtree(self.test_dir.parent, ignore_errors=True)

    async def start_all(self) -> bool:
        # Start server
        await self.start_server(self.server)
        await self.wait_for_server(self.server)

        # Start clients
        await asyncio.gather(*[self.start_client(c) for c in self.clients])
        await self.wait_for_clients(self.clients)

        return True

    async def cleanup(self):
        for process in self.__procs:
            if process.returncode is None:
                process.kill()
                await process.wait()
        for path in Path(".e2e").rglob("*.pid"):
            try:
                path.unlink()
            except Exception:
                pass

    async def start_server(self, server: Server) -> Process:
        server_dir = self.test_dir / "server"
        server_dir.mkdir(parents=True, exist_ok=True)

        logs_dir = self.test_dir / "logs"
        logs_dir.mkdir(parents=True, exist_ok=True)

        env = os.environ.copy()
        env["SYFTBOX_DATA_FOLDER"] = str(server_dir)
        env.update(server.env)

        process = await asyncio.create_subprocess_exec(
            "syftbox",
            "server",
            "--port",
            f"{server.port}",
            stdout=open(logs_dir / "server.log", "w"),
            stderr=asyncio.subprocess.STDOUT,
            env=env,
        )

        self.__procs.append(process)

        return process

    async def start_client(self, client: Client) -> Process:
        client_dir = self.test_dir / "clients" / client.name
        client_dir.mkdir(parents=True, exist_ok=True)
        client.data_dir = client_dir

        logs_dir = self.test_dir / "logs"
        logs_dir.mkdir(parents=True, exist_ok=True)

        env = os.environ.copy()
        env["SYFTBOX_DISABLE_ICONS"] = "1"
        env.update(client.env)
        if len(client.apps) > 0:
            env["SYFTBOX_DEFAULT_APPS"] = ",".join(client.apps)

        process = await asyncio.create_subprocess_exec(
            "syftbox",
            "client",
            f"--config={client_dir}/config.json",
            f"--data-dir={client_dir}",
            f"--email={client.name}@openmined.org",
            f"--server=http://localhost:{client.server_port}",
            f"--port={client.port}",
            "--no-open-dir",
            "--verbose",
            stdout=open(logs_dir / f"client.{client.name}.log", "w"),
            stderr=asyncio.subprocess.STDOUT,
            env=env,
        )
        self.__procs.append(process)
        return process

    async def wait_for_server(self, server: Server, timeout: int = 30):
        logger.debug(f"Waiting for server to be ready on port {server.port} (timeout={timeout}s)")
        await self.wait_for_url(f"http://localhost:{server.port}/info", timeout=timeout)
        logger.success(f"Server '{server.port}' is ready")

    async def wait_for_clients(self, clients: List[Client], timeout: int = 30):
        # wait for all futures with a timeout
        await asyncio.gather(*[self.wait_for_client(c, timeout=timeout) for c in clients])

    async def wait_for_client(self, client: Client, timeout: int = 30):
        logger.debug(f"Waiting for client '{client.name}' to be ready on port {client.port} (timeout={timeout}s)")
        await self.wait_for_url(f"http://localhost:{client.port}/info", timeout=timeout)
        logger.success(f"Client '{client.name}' is ready")

    async def wait_for_api(self, app_name: str, client: Client, timeout: int = 30):
        logger.debug(f"Waiting for API '{app_name}' to be ready (timeout={timeout}s)")
        run_path = client.api_dir / app_name / "run.sh"
        await self.wait_for_path(run_path, timeout=timeout)
        logger.success(f"API '{app_name}' is ready on client '{client.name}'")

    async def wait_for_path(self, path: Path, timeout: int = 30, interval: float = 0.5) -> None:
        logger.debug(f"Waiting for path '{path}' (timeout={timeout}s)")
        start = asyncio.get_event_loop().time()

        while not path.exists():
            if asyncio.get_event_loop().time() - start > timeout:
                raise E2ETimeoutError(f"Timeout after {timeout}s waiting for: {path}")
            await asyncio.sleep(interval)

        elapsed = asyncio.get_event_loop().time() - start
        logger.debug(f"Got {path} (after {elapsed:.1f}s)")

    async def wait_for_url(self, url: str, timeout: int = 30, interval: float = 1.0) -> None:
        async with httpx.AsyncClient() as client:
            start = asyncio.get_event_loop().time()
            while True:
                try:
                    await client.get(url)
                    break
                except httpx.RequestError:
                    if asyncio.get_event_loop().time() - start > timeout:
                        raise E2ETimeoutError(f"Timeout after {timeout}s waiting for: {url}")
                    await asyncio.sleep(interval)

            elapsed = asyncio.get_event_loop().time() - start
            logger.debug(f"Got response from {url} (after {elapsed:.1f}s)")


def get_random_port():
    import socket

    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("localhost", 0))
        return s.getsockname()[1]


@pytest_asyncio.fixture
async def e2e_context(request):
    try:
        ctx = E2EContext(**request.param)
        yield ctx
    except:
        raise
    finally:
        await ctx.cleanup()



from pathlib import Path
from typing import Optional

import pathspec
from loguru import logger

from syftbox.lib.types import PathLike, to_path

IGNORE_FILENAME = "_.syftignore"

DEFAULT_IGNORE = """
# Syft
/_.syftignore
/.syft*
/apps
/staging
/syft_changelog

# Python
.ipynb_checkpoints/
__pycache__/
*.py[cod]
.venv/

# OS-specific
.DS_Store
Icon

# IDE/Editor-specific
*.swp
*.swo
.vscode/
.idea/
*.iml

# General excludes
*.tmp

# excluded datasites
# example:
# /user_to_exclude@example.com/
"""


def create_default_ignore_file(dir: Path) -> None:
    """Create a default _.syftignore file in the dir"""
    ignore_file = to_path(dir) / IGNORE_FILENAME
    if not ignore_file.is_file():
        logger.info(f"Creating default ignore file: {ignore_file}")
        ignore_file.parent.mkdir(parents=True, exist_ok=True)
        ignore_file.write_text(DEFAULT_IGNORE)


def get_ignore_rules(dir: Path) -> Optional[pathspec.PathSpec]:
    """Get the ignore rules from the _.syftignore file in the dir"""
    ignore_file = to_path(dir) / IGNORE_FILENAME
    if ignore_file.is_file():
        with open(ignore_file) as f:
            lines = f.readlines()
        return pathspec.PathSpec.from_lines("gitwildmatch", lines)
    return None


def is_within_symlinked_path(path: Path, datasites_dir: PathLike) -> bool:
    """
    Returns True if the path is within a symlinked path.

    Symlinks are checked up to the datasites_dir.
    """
    base_dir = to_path(datasites_dir)
    for parent in path.parents:
        if parent == base_dir:
            break
        if parent.is_symlink():
            return True
    return False


def filter_symlinks(datasites_dir: Path, relative_paths: list[Path]) -> list[Path]:
    result = []
    for path in relative_paths:
        abs_path = datasites_dir / path

        is_symlinked = abs_path.is_symlink() or is_within_symlinked_path(abs_path, datasites_dir)
        if not is_symlinked:
            result.append(path)
    return result


def filter_hidden_files(relative_paths: list[Path]) -> list[Path]:
    result = []
    for path in relative_paths:
        if not any(part.startswith(".") for part in path.parts):
            result.append(path)
    return result


def filter_ignored_paths(
    datasites_dir: Path,
    relative_paths: list[Path],
    ignore_hidden_files: bool = True,
    ignore_symlinks: bool = True,
) -> list[Path]:
    """
    Filter out paths that are ignored. Ignore rules:
    - By default hidden files, or files within hidden directories are ignored.
    - By default symlinks are ignored, or files within symlinked directories are ignored.
    - files that match the ignore rules in the _.syftignore file are ignored.

    Args:
        datasites_dir (Path): Directory containing datasites.
        relative_paths (list[Path]): List of relative paths to filter. Paths are relative to datasites_dir.
        ignore_hidden_files (bool, optional): If True, all hidden files and directories are filtered. Defaults to True.
        ignore_symlinks (bool, optional): if True, all symlinked files and folders are filtered. Defaults to True.

    Returns:
        list[Path]: List of filtered relative paths.
    """

    if ignore_hidden_files:
        relative_paths = filter_hidden_files(relative_paths)

    if ignore_symlinks:
        relative_paths = filter_symlinks(datasites_dir, relative_paths)

    ignore_rules = get_ignore_rules(datasites_dir)
    if ignore_rules is None:
        return relative_paths

    filtered_paths = []
    for path in relative_paths:
        if not ignore_rules.match_file(path):
            filtered_paths.append(path)

    return filtered_paths



from secrets import token_hex

from syftbox.client.cli_setup import setup_config_interactive
from syftbox.lib.client_config import SyftClientConfig


def test_setup_new_config(tmp_path):
    config_path = tmp_path / f"config_{token_hex}.json"
    data_dir = tmp_path / "data"
    email = "test@example.com"
    server = "http://test.com/"
    port = 8080

    result = setup_config_interactive(
        config_path=config_path, email=email, data_dir=data_dir, server=server, port=port, skip_auth=True
    )

    assert isinstance(result, SyftClientConfig)
    assert result.path == config_path
    assert result.data_dir == data_dir
    assert str(result.email) == str(email)
    assert str(result.server_url) == str(server)
    assert str(result.client_url) == "http://127.0.0.1:8080/"
    assert result.client_url.port == port


def test_setup_new_config_with_prompt(tmp_path, monkeypatch):
    config_path = tmp_path / f"config_{token_hex}.json"
    data_dir = tmp_path / "data"
    email = "test@example.com"
    server = "http://test.com/"
    port = 8080

    monkeypatch.setattr("syftbox.client.cli_setup.prompt_email", lambda: email)
    monkeypatch.setattr("syftbox.client.cli_setup.prompt_data_dir", lambda: data_dir)

    result = setup_config_interactive(
        config_path=config_path,
        email=None,
        data_dir=None,
        server=server,
        port=port,
        skip_auth=True,
    )

    assert isinstance(result, SyftClientConfig)
    assert result.path == config_path
    assert result.data_dir == data_dir
    assert str(result.email) == str(email)
    assert str(result.server_url) == str(server)
    assert str(result.client_url) == "http://127.0.0.1:8080/"
    assert result.client_url.port == port


def test_setup_existing_config(tmp_path, mock_config):
    new_port = 8081
    result = setup_config_interactive(
        server=str(mock_config.server_url),
        email=str(mock_config.email),
        data_dir="",
        config_path=mock_config.path,
        port=new_port,
        skip_auth=True,
    )

    assert isinstance(result, SyftClientConfig)
    assert result.client_url.port == new_port



from pathlib import Path

import syftbox
from syftbox.client.plugins.apps import run_apps


class AppFixture:
    @staticmethod
    def create_app_with_name_check(apps_dir: Path, app_name: str) -> tuple[Path, str]:
        """Create a test app that outputs its api_request_name"""
        app_dir = apps_dir / app_name
        app_dir.mkdir(parents=True, exist_ok=True)

        # Create a Python script that uses api_request_name
        test_script = app_dir / "main.py"
        test_script.write_text("""
from syftbox.lib import Client

client = Client.load()
print(f"API_NAME:{client.api_request_name}")
""")

        # Create runner script
        SYFTBOX_SOURCE_PATH = Path(syftbox.__file__).parent.parent
        run_script = app_dir / "run.sh"
        run_script.write_text(f"""#!/bin/bash
set -e

uv venv && . .venv/bin/activate
uv pip install --upgrade --editable {SYFTBOX_SOURCE_PATH}
python3 main.py | tee app.log
deactivate
""")
        run_script.chmod(0o755)
        return app_dir


def test_api_request_name_in_app(tmp_path, mock_config):
    """Test that api_request_name returns correct name when called from within an app"""
    # Setup
    app_name = "test_app_that_echoes_name"
    app_dir = AppFixture.create_app_with_name_check(tmp_path, app_name)

    # Run the app
    run_apps(tmp_path, mock_config.path)

    # Verify the output
    app_log = app_dir / "app.log"
    assert app_log.exists()
    output = app_log.read_text().strip()
    assert f"API_NAME:{app_name}" in output



import base64
import hashlib
import sqlite3
import zipfile
from io import BytesIO

import py_fast_rsync
from fastapi import APIRouter, Depends, HTTPException, Request, Response, UploadFile
from fastapi.responses import FileResponse, JSONResponse, StreamingResponse
from loguru import logger

from syftbox.lib.lib import PermissionTree, SyftPermission, filter_metadata
from syftbox.server.analytics import log_file_change_event
from syftbox.server.settings import ServerSettings, get_server_settings
from syftbox.server.sync.db import (
    get_all_datasites,
    get_db,
)
from syftbox.server.sync.file_store import FileStore, SyftFile
from syftbox.server.users.auth import get_current_user

from .models import (
    ApplyDiffRequest,
    ApplyDiffResponse,
    BatchFileRequest,
    DiffRequest,
    DiffResponse,
    FileMetadata,
    FileMetadataRequest,
    FileRequest,
    RelativePath,
)


def get_db_connection(request: Request):
    conn = get_db(request.state.server_settings.file_db_path)
    yield conn
    conn.close()


def get_file_store(request: Request):
    store = FileStore(
        server_settings=request.state.server_settings,
    )
    yield store


router = APIRouter(prefix="/sync", tags=["sync"])


@router.post("/get_diff", response_model=DiffResponse)
def get_diff(
    req: DiffRequest,
    file_store: FileStore = Depends(get_file_store),
    email: str = Depends(get_current_user),
) -> DiffResponse:
    try:
        file = file_store.get(req.path)
    except ValueError:
        raise HTTPException(status_code=404, detail="file not found")
    diff = py_fast_rsync.diff(req.signature_bytes, file.data)
    diff_bytes = base64.b85encode(diff).decode("utf-8")
    return DiffResponse(
        path=file.metadata.path.as_posix(),
        diff=diff_bytes,
        hash=file.metadata.hash,
    )


@router.post("/datasite_states", response_model=dict[str, list[FileMetadata]])
def get_datasite_states(
    conn: sqlite3.Connection = Depends(get_db_connection),
    file_store: FileStore = Depends(get_file_store),
    server_settings: ServerSettings = Depends(get_server_settings),
    email: str = Depends(get_current_user),
) -> dict[str, list[FileMetadata]]:
    all_datasites = get_all_datasites(conn)
    datasite_states: dict[str, list[FileMetadata]] = {}
    for datasite in all_datasites:
        try:
            datasite_state = dir_state(RelativePath(datasite), file_store, server_settings, email)
        except Exception as e:
            logger.error(f"Failed to get dir state for {datasite}: {e}")
            continue
        datasite_states[datasite] = datasite_state

    return datasite_states


@router.post("/dir_state", response_model=list[FileMetadata])
def dir_state(
    dir: RelativePath,
    file_store: FileStore = Depends(get_file_store),
    server_settings: ServerSettings = Depends(get_server_settings),
    email: str = Depends(get_current_user),
) -> list[FileMetadata]:
    full_path = server_settings.snapshot_folder / dir
    # get the top level perm file
    try:
        perm_tree = PermissionTree.from_path(full_path, raise_on_corrupted_files=True)
    except ValueError:
        raise HTTPException(status_code=500, detail=f"Failed to parse permission tree: {dir}")

    # filter the read state for this user by the perm tree
    metadata_list = file_store.list(dir)
    filtered_metadata = filter_metadata(email, metadata_list, perm_tree, server_settings.snapshot_folder)
    return filtered_metadata


@router.post("/get_metadata", response_model=FileMetadata)
def get_metadata(
    req: FileMetadataRequest,
    file_store: FileStore = Depends(get_file_store),
    email: str = Depends(get_current_user),
) -> FileMetadata:
    try:
        metadata = file_store.get_metadata(req.path_like)
        return metadata
    except ValueError as e:
        raise HTTPException(status_code=400, detail=str(e))


@router.post("/apply_diff", response_model=ApplyDiffResponse)
def apply_diffs(
    req: ApplyDiffRequest,
    file_store: FileStore = Depends(get_file_store),
    email: str = Depends(get_current_user),
) -> ApplyDiffResponse:
    try:
        file = file_store.get(req.path)
    except ValueError:
        raise HTTPException(status_code=404, detail="file not found")

    result = py_fast_rsync.apply(file.data, req.diff_bytes)
    new_hash = hashlib.sha256(result).hexdigest()

    if new_hash != req.expected_hash:
        raise HTTPException(status_code=400, detail="hash mismatch, skipped writing")

    if SyftPermission.is_permission_file(file.metadata.path) and not SyftPermission.is_valid(result):
        raise HTTPException(status_code=400, detail="invalid syftpermission contents, skipped writing")

    file_store.put(req.path, result)

    log_file_change_event(
        "/sync/apply_diff",
        email=email,
        relative_path=req.path,
        file_store=file_store,
    )

    return ApplyDiffResponse(path=req.path, current_hash=new_hash, previous_hash=file.metadata.hash)


@router.post("/delete", response_class=JSONResponse)
def delete_file(
    req: FileRequest,
    file_store: FileStore = Depends(get_file_store),
    email: str = Depends(get_current_user),
) -> JSONResponse:
    log_file_change_event(
        "/sync/delete",
        email=email,
        relative_path=req.path,
        file_store=file_store,
    )

    file_store.delete(req.path)
    return JSONResponse(content={"status": "success"})


@router.post("/create", response_class=JSONResponse)
def create_file(
    file: UploadFile,
    file_store: FileStore = Depends(get_file_store),
    email: str = Depends(get_current_user),
) -> JSONResponse:
    relative_path = RelativePath(file.filename)
    if "%" in file.filename:
        raise HTTPException(status_code=400, detail="filename cannot contain '%'")

    if file_store.exists(relative_path):
        raise HTTPException(status_code=400, detail="file already exists")

    contents = file.file.read()

    if SyftPermission.is_permission_file(relative_path) and not SyftPermission.is_valid(contents):
        raise HTTPException(status_code=400, detail="invalid syftpermission contents, skipped writing")

    file_store.put(
        relative_path,
        contents,
    )

    log_file_change_event(
        "/sync/create",
        email=email,
        relative_path=relative_path,
        file_store=file_store,
    )
    return JSONResponse(content={"status": "success"})


@router.post("/download", response_class=FileResponse)
def download_file(
    req: FileRequest,
    file_store: FileStore = Depends(get_file_store),
    email: str = Depends(get_current_user),
) -> FileResponse:
    try:
        abs_path = file_store.get(req.path).absolute_path
        return FileResponse(abs_path)
    except ValueError as e:
        raise HTTPException(status_code=400, detail=str(e))


@router.post("/datasites", response_model=list[str])
def get_datasites(
    conn: sqlite3.Connection = Depends(get_db_connection),
    email: str = Depends(get_current_user),
) -> list[str]:
    return get_all_datasites(conn)


def create_zip_from_files(files: list[SyftFile]) -> BytesIO:
    memory_file = BytesIO()
    with zipfile.ZipFile(memory_file, "w") as zf:
        for file in files:
            zf.writestr(file.metadata.path.as_posix(), file.data)
    memory_file.seek(0)
    return memory_file


@router.post("/download_bulk")
async def get_files(
    req: BatchFileRequest,
    file_store: FileStore = Depends(get_file_store),
    email: str = Depends(get_current_user),
) -> StreamingResponse:
    all_files = []
    for path in req.paths:
        try:
            file = file_store.get(path)
        except ValueError:
            logger.warning(f"File not found: {path}")
            continue
        all_files.append(file)
    zip_file = create_zip_from_files(all_files)
    return Response(content=zip_file.read(), media_type="application/zip")



from typing import Optional

from syftbox.lib.exceptions import SyftBoxException


class SyftInitializationError(SyftBoxException):
    pass


class SyftBoxAlreadyRunning(SyftBoxException):
    pass


class SyftServerError(SyftBoxException):
    pass


class SyftAuthenticationError(SyftServerError):
    default_message = "Authentication failed, please log in again."

    def __init__(self, message: Optional[str] = None):
        message = self.default_message if message is None else message
        super().__init__(message)


class SyftNotFound(SyftServerError):
    pass



import base64
import hashlib
import re
from concurrent.futures import ProcessPoolExecutor
from datetime import datetime, timezone
from functools import partial
from pathlib import Path
from typing import Optional, Union

from loguru import logger
from py_fast_rsync import signature

from syftbox.lib.ignore import filter_ignored_paths
from syftbox.server.sync.models import FileMetadata


def hash_file(file_path: Path, root_dir: Optional[Path] = None) -> Optional[FileMetadata]:
    # ignore files larger then 100MB
    try:
        if file_path.stat().st_size > 100_000_000:
            logger.warning("File too large: %s", file_path)
            return str(file_path), None

        with open(file_path, "rb") as f:
            # not ideal for large files
            # but py_fast_rsync does not support files yet.
            # TODO: add support for streaming hashing
            data = f.read()

        if root_dir is None:
            path = file_path
        else:
            path = file_path.relative_to(root_dir)
        return FileMetadata(
            path=path,
            hash=hashlib.sha256(data).hexdigest(),
            signature=base64.b85encode(signature.calculate(data)),
            file_size=len(data),
            last_modified=datetime.fromtimestamp(file_path.stat().st_mtime, timezone.utc),
        )
    except Exception:
        logger.error(f"Failed to hash file {file_path}")
        return None


def hash_files_parallel(files: list[Path], root_dir: Path) -> list[FileMetadata]:
    with ProcessPoolExecutor() as executor:
        results = list(executor.map(partial(hash_file, root_dir=root_dir), files))
    return [r for r in results if r is not None]


def hash_files(files: list[Path], root_dir: Path) -> list[FileMetadata]:
    result = [hash_file(file, root_dir) for file in files]
    return [r for r in result if r is not None]


def hash_dir(
    dir: Path,
    root_dir: Path,
    filter_ignored: bool = True,
) -> list[FileMetadata]:
    """
    hash all files in dir recursively, return a list of FileMetadata.

    ignore_folders should be relative to root_dir.
    returned Paths are relative to root_dir.
    """
    files = collect_files(dir)

    relative_paths = [file.relative_to(root_dir) for file in files]
    if filter_ignored:
        relative_paths = filter_ignored_paths(root_dir, relative_paths)

    absolute_paths = [root_dir / file for file in relative_paths]
    return hash_files(absolute_paths, root_dir)


def collect_files(
    dir: Union[Path, str],
    pattern: Union[str, re.Pattern, None] = None,
) -> list[Path]:
    """Recursively collect files in a directory with options to include hidden files and symlinks"""

    dir = Path(dir)
    if not dir.is_dir():
        return []
    files = []

    # Compile the regex pattern if it's a string
    if isinstance(pattern, str):
        pattern = re.compile(pattern)

    for entry in dir.iterdir():
        if entry.is_file():
            if pattern is None or pattern.match(entry.as_posix()):
                files.append(entry)
        elif entry.is_dir():
            files.extend(collect_files(entry, pattern))

    return files



import json
import os
import shutil
from pathlib import Path
from typing import Optional

from pydantic import AliasChoices, AnyHttpUrl, BaseModel, ConfigDict, EmailStr, Field, field_validator
from pydantic_core import Url
from typing_extensions import Self

from syftbox.lib.constants import DEFAULT_CONFIG_PATH, DEFAULT_DATA_DIR, DEFAULT_SERVER_URL
from syftbox.lib.exceptions import ClientConfigException
from syftbox.lib.types import PathLike, to_path

__all__ = ["SyftClientConfig"]

# env or default
CONFIG_PATH_ENV = "SYFTBOX_CLIENT_CONFIG_PATH"

# Old configuration file path for the client
LEGACY_CONFIG_NAME = "client_config.json"


class SyftClientConfig(BaseModel):
    """SyftBox client configuration"""

    # model config
    model_config = ConfigDict(extra="ignore", json_encoders={AnyHttpUrl: lambda v: str(v)})

    data_dir: Path = Field(
        validation_alias=AliasChoices("data_dir", "sync_folder"),
        default=DEFAULT_DATA_DIR,
        description="Local directory where client data is stored",
    )
    """Local directory where client data is stored"""

    server_url: AnyHttpUrl = Field(
        default=DEFAULT_SERVER_URL,
        description="URL of the remote SyftBox server",
    )
    """URL of the remote SyftBox server"""

    client_url: AnyHttpUrl = Field(
        validation_alias=AliasChoices("client_url", "port"),
        description="URL where the client is running",
    )
    """URL where the client is running"""

    email: EmailStr = Field(description="Email address of the user")
    """Email address of the user"""

    token: Optional[str] = Field(
        default=None, description="Depracated: Use access_token instead. API token for the user", deprecated=True
    )
    """Depracated: Use access_token instead. API token for the user"""

    access_token: Optional[str] = Field(default=None, description="Access token for the user")
    """Access token for the user"""

    # WARN: we don't need `path` to be serialized, hence exclude=True
    path: Path = Field(exclude=True, description="Path to the config file")
    """Path to the config file"""

    @field_validator("client_url", mode="before")
    def port_to_url(cls, val):
        if isinstance(val, int):
            return f"http://127.0.0.1:{val}"
        return val

    @field_validator("token", mode="before")
    def token_to_str(cls, v):
        if not v:
            return None
        elif isinstance(v, int):
            return str(v)
        return v

    def set_server_url(self, server: str):
        self.server_url = Url(server)

    def set_port(self, port: int):
        self.client_url = Url(f"http://127.0.0.1:{port}")

    @classmethod
    def load(cls, conf_path: Optional[PathLike] = None) -> Self:
        try:
            # args or env or default
            path = conf_path or os.getenv(CONFIG_PATH_ENV, DEFAULT_CONFIG_PATH)
            path = to_path(path)

            # todo migration stuff we can remove later
            legacy_path = Path(path.parent, LEGACY_CONFIG_NAME)
            # prefer to load config.json instead of client_config.json
            # initially config.json WILL NOT exist, so we fallback to client_config.json
            if path.exists():
                data = json.loads(path.read_text())
            elif legacy_path.exists():
                data = json.loads(legacy_path.read_text())
                path = legacy_path
            # todo end

            return cls(path=path, **data)
        except Exception as e:
            raise ClientConfigException(f"Failed to load config from '{conf_path}' - {e}")

    @classmethod
    def exists(cls, path: PathLike) -> bool:
        return to_path(path).exists()

    def migrate(self) -> Self:
        """Explicit call to migrate the config file"""

        # if we loaded the legacy config, we need to move it to new config
        if self.path.name == LEGACY_CONFIG_NAME:
            new_path = Path(self.path.parent, DEFAULT_CONFIG_PATH.name)
            shutil.move(str(self.path), str(new_path))
            self.path = new_path
            self.save()

        return self

    def as_dict(self, exclude=None):
        return self.model_dump(exclude=exclude, exclude_none=True, warnings="none")

    def as_json(self, indent=4):
        return self.model_dump_json(indent=indent, exclude_none=True, warnings="none")

    def save(self) -> Self:
        self.path.parent.mkdir(parents=True, exist_ok=True)
        self.path.write_text(self.as_json())
        return self



import json
import os
import subprocess
from pathlib import Path

import yaml
from fastapi import APIRouter, HTTPException
from fastapi.responses import JSONResponse
from pydantic import BaseModel
from typing_extensions import List

from syftbox.client.routers.common import APIContext

router = APIRouter()


def parse_frontmatter(file_path):
    """
    Parses frontmatter YAML from a README.md file and returns it as a Python dictionary.

    Args:
        file_path (str): Path to the README.md file.

    Returns:
        dict: The parsed YAML frontmatter as a dictionary. If no frontmatter is found, returns an empty dict.
    """
    with open(file_path, "r", encoding="utf-8") as file:
        lines = file.readlines()

    # Check for YAML frontmatter boundaries
    if lines[0].strip() == "---":
        yaml_lines = []
        for line in lines[1:]:
            if line.strip() == "---":
                break
            yaml_lines.append(line)

        # Parse the YAML content
        frontmatter = yaml.safe_load("".join(yaml_lines))
        return frontmatter if frontmatter else {}
    else:
        return {}


class AppDetails(BaseModel):
    name: str
    version: str
    source: str
    home: str
    icon: str
    path: str


def get_all_apps(apps_dir: str) -> List[AppDetails]:
    """
    Get all apps in the given directory.

    Args:
        apps_dir (str): Path to the directory containing the apps.

    Returns:
        list: A list of AppDetails objects.
    """
    apps = []
    for app_dir in Path(apps_dir).iterdir():
        if app_dir.is_dir():
            readme_path = app_dir / "README.md"
            if readme_path.exists():
                frontmatter = parse_frontmatter(readme_path)
                app = AppDetails(
                    name=frontmatter.get("name", app_dir.name),
                    version=frontmatter.get("version", "0.0.1"),
                    source=frontmatter.get("source", ""),
                    home=frontmatter.get("home", ""),
                    icon=frontmatter.get("icon", ""),
                    path=str(app_dir),
                )
                apps.append(app)

    return apps


@router.get("/")
async def index(ctx: APIContext):
    apps_dir = ctx.workspace.apps
    apps = get_all_apps(apps_dir)

    return JSONResponse(content=[app.model_dump() for app in apps])


@router.get("/status/{app_name}")
async def app_details(ctx: APIContext, app_name: str):
    apps_dir = ctx.workspace.apps
    apps = get_all_apps(apps_dir)
    for app in apps:
        if app_name == app.name:
            return JSONResponse(content=app.model_dump())
    return JSONResponse(status_code=404, content={"message": "App not found"})


class InstallRequest(BaseModel):
    source: str
    version: str


@router.post("/install")
async def install_app(request: InstallRequest):
    command = ["syftbox", "app", "install", request.source, "--called-by", "api"]
    try:
        # Run the command and capture output
        result = subprocess.run(command, capture_output=True, text=True, check=True)
        print(result.stderr, result.stdout)
        # If successful, return JSON payload indicating success
        return {
            "status": "success",
            "message": f"App {request.source} version {request.version} installed successfully.",
            "output": result.stdout,
        }

    except subprocess.CalledProcessError as e:
        # Handle command failure, return JSON with error details
        raise HTTPException(
            status_code=500,
            detail={"status": "error", "message": f"Failed to install app {request.source}.", "output": e.stderr},
        )


@router.post("/command/{app_name}")
async def app_command(ctx: APIContext, app_name: str, request: dict):
    apps_dir = ctx.workspace.apps
    apps = get_all_apps(apps_dir)

    for app in apps:
        if app_name == app.name:
            # Convert request dictionary to JSON string and wrap with single quotes for shell
            request_json = json.dumps(request)
            json_arg = f"--input='{request_json}'"  # Wrap entire JSON argument in single quotes
            command = f"uv run {app.path}/command.py {json_arg}"  # Complete command as a single string
            print("command", command)

            # Define the environment variable
            env = {
                **os.environ,
                "SYFTBOX_CLIENT_CONFIG_PATH": ctx.config.path,
            }

            try:
                # Execute the command with the specified environment
                result = subprocess.run(command, check=True, capture_output=True, text=True, shell=True, env=env)

                # Trim the output and attempt to parse it as JSON
                trimmed_output = result.stdout.strip()
                try:
                    json_output = json.loads(trimmed_output)
                    return JSONResponse(content=json_output)
                except json.JSONDecodeError:
                    # Return trimmed output as plain text if not valid JSON
                    return JSONResponse(content={"output": trimmed_output})
            except subprocess.CalledProcessError as e:
                print("error", e)
                return JSONResponse(status_code=500, content={"error": e.stderr.strip()})

    raise HTTPException(status_code=404, detail="App not found")


@router.get("/logs/{app_name}")
async def app_logs(ctx: APIContext, app_name: str):
    apps_dir = ctx.workspace.apps
    all_apps = get_all_apps(apps_dir)
    app_details = None
    for app in all_apps:
        if app_name == app.name:
            app_details = app

    # Raise 404 if app not found
    if app_details is None:
        raise HTTPException(status_code=404, detail="App not found")

    logs = []

    # Read the log file if it exists
    app_path = Path(app_details.path)

    log_file = app_path / "logs" / f"{app_name}.log"
    if log_file.exists():
        with open(log_file, "r") as file:
            logs = file.readlines()

    return JSONResponse(content={"logs": logs})



import os
import shutil
import subprocess
from pathlib import Path

DEFAULT_APPS = [
    "https://github.com/OpenMined/logged_in",
    "https://github.com/OpenMined/inbox",
]


def clone_apps():
    apps = DEFAULT_APPS

    # this is needed for E2E or integration testing to only install only select apps
    # DO NOT MERGE IT WITH DEFAULT_APPS
    env_apps = os.getenv("SYFTBOX_DEFAULT_APPS", None)
    if env_apps:
        print(f"SYFTBOX_DEFAULT_APPS={env_apps}")
        apps = env_apps.strip().split(",")

    print("Installing", apps)

    # Iterate over the list and clone each repository
    for url in apps:
        subprocess.run(["git", "clone", url])

    print("Done")


if __name__ == "__main__":
    current_directory = Path(os.getcwd())

    apps_directory = current_directory.parent
    os.chdir(apps_directory)
    clone_apps()
    shutil.rmtree(current_directory)



import contextlib
import json
import os
import platform
import random
import sys
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path

from fastapi import Depends, FastAPI, Request
from fastapi.middleware.gzip import GZipMiddleware
from fastapi.responses import (
    FileResponse,
    HTMLResponse,
    JSONResponse,
    PlainTextResponse,
    RedirectResponse,
)
from jinja2 import Template
from loguru import logger
from typing_extensions import Any, Optional, Union

from syftbox.__version__ import __version__
from syftbox.lib.lib import (
    Jsonable,
    get_datasites,
)
from syftbox.server.analytics import log_analytics_event
from syftbox.server.logger import setup_logger
from syftbox.server.middleware import LoguruMiddleware
from syftbox.server.settings import ServerSettings, get_server_settings
from syftbox.server.users.auth import get_current_user

from .emails.router import router as emails_router
from .sync import db, hash
from .sync.router import router as sync_router
from .users.router import router as users_router

current_dir = Path(__file__).parent


def load_dict(cls, filepath: str) -> Optional[dict[str, Any]]:
    try:
        with open(filepath) as f:
            data = f.read()
            d = json.loads(data)
            dicts = {}
            for key, value in d.items():
                dicts[key] = cls(**value)
            return dicts
    except Exception as e:
        logger.info(f"Unable to load dict file: {filepath}. {e}")
    return None


def save_dict(obj: Any, filepath: str) -> None:
    dicts = {}
    for key, value in obj.items():
        dicts[key] = value.to_dict()

    with open(filepath, "w") as f:
        f.write(json.dumps(dicts))


@dataclass
class User(Jsonable):
    email: str
    token: int  # TODO


class Users:
    def __init__(self, path: Path) -> None:
        self.path = path
        self.users = {}
        self.load()

    def load(self):
        if os.path.exists(str(self.path)):
            users = load_dict(User, str(self.path))
        else:
            users = None
        if users:
            self.users = users

    def save(self):
        save_dict(self.users, str(self.path))

    def get_user(self, email: str) -> Optional[User]:
        if email not in self.users:
            return None
        return self.users[email]

    def create_user(self, email: str) -> int:
        if email in self.users:
            # for now just return the token
            return self.users[email].token
            # raise Exception(f"User already registered: {email}")
        token = random.randint(0, sys.maxsize)
        user = User(email=email, token=token)
        self.users[email] = user
        self.save()
        return token

    def __repr__(self) -> str:
        string = ""
        for email, user in self.users.items():
            string += f"{email}: {user}"
        return string


def get_users(request: Request) -> Users:
    return request.state.users


def create_folders(folders: list[str]) -> None:
    for folder in folders:
        if not os.path.exists(folder):
            os.makedirs(folder, exist_ok=True)


def init_db(settings: ServerSettings) -> None:
    # might take very long as snapshot folder grows
    logger.info(f"> Collecting Files from {settings.snapshot_folder.absolute()}")
    files = hash.collect_files(settings.snapshot_folder.absolute())
    logger.info("> Hashing files")
    metadata = hash.hash_files(files, settings.snapshot_folder)
    logger.info(f"> Updating file hashes at {settings.file_db_path.absolute()}")
    con = db.get_db(settings.file_db_path.absolute())
    cur = con.cursor()
    for m in metadata:
        db.save_file_metadata(cur, m)

    cur.close()
    con.commit()
    con.close()


@contextlib.asynccontextmanager
async def lifespan(app: FastAPI, settings: Optional[ServerSettings] = None):
    # Startup
    if settings is None:
        settings = ServerSettings()

    setup_logger(logs_folder=settings.logs_folder)

    logger.info(f"> Starting SyftBox Server {__version__}. Python {platform.python_version()}")
    logger.info(settings)

    logger.info("> Creating Folders")

    create_folders(settings.folders)

    users = Users(path=settings.user_file_path)
    logger.info("> Loading Users")
    logger.info(users)

    init_db(settings)

    yield {
        "server_settings": settings,
        "users": users,
    }

    logger.info("> Shutting down server")


app = FastAPI(lifespan=lifespan)
app.include_router(emails_router)
app.include_router(sync_router)
app.include_router(users_router)
app.add_middleware(GZipMiddleware, minimum_size=1000, compresslevel=5)
app.add_middleware(LoguruMiddleware)

# Define the ASCII art
ascii_art = rf"""
 ____         __ _   ____
/ ___| _   _ / _| |_| __ )  _____  __
\___ \| | | | |_| __|  _ \ / _ \ \/ /
 ___) | |_| |  _| |_| |_) | (_) >  <
|____/ \__, |_|  \__|____/ \___/_/\_\
       |___/        {__version__:>17}


# Install Syftbox (MacOS and Linux)
curl -LsSf https://syftbox.openmined.org/install.sh | sh

# Run the client
syftbox client
"""


@app.get("/", response_class=PlainTextResponse)
async def get_ascii_art(request: Request):
    req_host = request.headers.get("host", "")
    if "syftboxstage" in req_host:
        return ascii_art.replace("syftbox.openmined.org", "syftboxstage.openmined.org")
    return ascii_art


@app.get("/wheel/{path:path}", response_class=HTMLResponse)
async def get_wheel(path: str):
    if path == "":  # Check if path is empty (meaning "/datasites/")
        return RedirectResponse(url="/")

    filename = path.split("/")[0]
    if filename.endswith(".whl"):
        wheel_path = os.path.expanduser("~/syftbox-0.1.0-py3-none-any.whl")
        return FileResponse(wheel_path, media_type="application/octet-stream")
    return filename


def get_file_list(directory: Union[str, Path] = ".") -> list[dict[str, Any]]:
    # TODO rewrite with pathlib
    directory = str(directory)

    file_list = []
    for item in os.listdir(directory):
        item_path = os.path.join(directory, item)
        is_dir = os.path.isdir(item_path)
        size = os.path.getsize(item_path) if not is_dir else "-"
        mod_time = datetime.fromtimestamp(os.path.getmtime(item_path)).strftime("%Y-%m-%d %H:%M:%S")

        file_list.append({"name": item, "is_dir": is_dir, "size": size, "mod_time": mod_time})

    return sorted(file_list, key=lambda x: (not x["is_dir"], x["name"].lower()))


@app.get("/datasites", response_class=HTMLResponse)
async def list_datasites(request: Request, server_settings: ServerSettings = Depends(get_server_settings)):
    files = get_file_list(server_settings.snapshot_folder)
    template_path = current_dir / "templates" / "datasites.html"
    html = ""
    with open(template_path) as f:
        html = f.read()
    template = Template(html)

    html_content = template.render(
        {
            "request": request,
            "files": files,
            "current_path": "/",
        }
    )
    return html_content


@app.get("/datasites/{path:path}", response_class=HTMLResponse)
async def browse_datasite(
    request: Request,
    path: str,
    server_settings: ServerSettings = Depends(get_server_settings),
):
    if path == "":  # Check if path is empty (meaning "/datasites/")
        return RedirectResponse(url="/datasites")

    snapshot_folder = str(server_settings.snapshot_folder)
    datasite_part = path.split("/")[0]
    datasites = get_datasites(snapshot_folder)
    if datasite_part in datasites:
        slug = path[len(datasite_part) :]
        if slug == "":
            slug = "/"
        datasite_path = os.path.join(snapshot_folder, datasite_part)
        datasite_public = datasite_path + "/public"
        if not os.path.exists(datasite_public):
            return "No public datasite"

        slug_path = os.path.abspath(datasite_public + slug)
        if os.path.exists(slug_path) and os.path.isfile(slug_path):
            if slug_path.endswith(".html") or slug_path.endswith(".htm"):
                return FileResponse(slug_path)
            elif slug_path.endswith(".md"):
                with open(slug_path, "r") as file:
                    content = file.read()
                return PlainTextResponse(content)
            elif slug_path.endswith(".json") or slug_path.endswith(".jsonl"):
                return FileResponse(slug_path, media_type="application/json")
            elif slug_path.endswith(".yaml") or slug_path.endswith(".yml"):
                return FileResponse(slug_path, media_type="application/x-yaml")
            elif slug_path.endswith(".log") or slug_path.endswith(".txt"):
                return FileResponse(slug_path, media_type="text/plain")
            elif slug_path.endswith(".py"):
                return FileResponse(slug_path, media_type="text/plain")
            else:
                return FileResponse(slug_path, media_type="application/octet-stream")

        # show directory
        if not path.endswith("/") and os.path.exists(path + "/") and os.path.isdir(path + "/"):
            return RedirectResponse(url=f"{path}/")

        index_file = os.path.abspath(slug_path + "/" + "index.html")
        if os.path.exists(index_file):
            with open(index_file, "r") as file:
                html_content = file.read()
            return HTMLResponse(content=html_content, status_code=200)

        if os.path.isdir(slug_path):
            files = get_file_list(slug_path)
            template_path = current_dir / "templates" / "folder.html"
            html = ""
            with open(template_path) as f:
                html = f.read()
            template = Template(html)
            html_content = template.render(
                {
                    "datasite": datasite_part,
                    "request": request,
                    "files": files,
                    "current_path": path,
                }
            )
            return html_content
        else:
            # return 404
            message_404 = f"No file or directory found at /datasites/{datasite_part}{slug}"
            return HTMLResponse(content=message_404, status_code=404)

    return f"No Datasite {datasite_part} exists"


@app.post("/register")
async def register(
    request: Request,
    users: Users = Depends(get_users),
    server_settings: ServerSettings = Depends(get_server_settings),
):
    data = await request.json()
    email = data["email"]
    token = users.create_user(email)

    # create datasite snapshot folder
    datasite_folder = Path(server_settings.snapshot_folder) / email
    os.makedirs(datasite_folder, exist_ok=True)

    logger.info(f"> {email} registering: {token}, snapshot folder: {datasite_folder}")
    log_analytics_event("/register", email)

    return JSONResponse({"status": "success", "token": token}, status_code=200)


@app.post("/log_event")
async def log_event(
    request: Request,
    email: str = Depends(get_current_user),
):
    data = await request.json()
    log_analytics_event("/log_event", email, **data)
    return JSONResponse({"status": "success"}, status_code=200)


@app.get("/install.sh")
async def install():
    install_script = current_dir / "templates" / "install.sh"
    return FileResponse(install_script, media_type="text/plain")


@app.get("/icon.png")
async def icon():
    icon_path = current_dir / "assets" / "icon.png"
    return FileResponse(icon_path, media_type="image/png")


@app.get("/info")
async def info():
    return {
        "version": __version__,
    }



from pathlib import Path

# Default port for the SyftBox client
DEFAULT_PORT = 8080

# Default SyftBox cache server URL for the client
DEFAULT_SERVER_URL = "https://syftbox.openmined.org"

# Default configuration directory for the client
DEFAULT_CONFIG_DIR = Path(Path.home(), ".syftbox")

# Default configuration file path for the client
DEFAULT_CONFIG_PATH = Path(DEFAULT_CONFIG_DIR, "config.json")

# Default logs directory for the client
DEFAULT_LOGS_DIR = Path(DEFAULT_CONFIG_DIR, "logs")

# Default data directory for the client
DEFAULT_DATA_DIR = Path(Path.home(), "SyftBox")

# Permissions file name
PERM_FILE = "_.syftperm"



from pathlib import Path

from pydantic import BaseModel

from syftbox.server.settings import ServerSettings
from syftbox.server.sync import db
from syftbox.server.sync.db import get_db
from syftbox.server.sync.hash import hash_file
from syftbox.server.sync.models import AbsolutePath, FileMetadata, RelativePath


class SyftFile(BaseModel):
    metadata: FileMetadata
    data: bytes
    absolute_path: AbsolutePath


class FileStore:
    def __init__(self, server_settings: ServerSettings) -> None:
        self.server_settings = server_settings

    @property
    def db_path(self) -> AbsolutePath:
        return self.server_settings.file_db_path

    def delete(self, path: RelativePath) -> None:
        conn = get_db(self.db_path)
        cursor = conn.cursor()
        cursor.execute("BEGIN IMMEDIATE;")
        try:
            db.delete_file_metadata(cursor, str(path))
        except ValueError:
            pass
        abs_path = self.server_settings.snapshot_folder / path
        abs_path.unlink(missing_ok=True)
        conn.commit()
        cursor.close()

    def get(self, path: RelativePath) -> SyftFile:
        with get_db(self.db_path) as conn:
            metadata = db.get_one_metadata(conn, path=str(path))
            abs_path = self.server_settings.snapshot_folder / metadata.path

            if not Path(abs_path).exists():
                self.delete(metadata.path.as_posix())
                raise ValueError("File not found")
            return SyftFile(metadata=metadata, data=self._read_bytes(abs_path), absolute_path=abs_path)

    def exists(self, path: RelativePath) -> bool:
        with get_db(self.db_path) as conn:
            try:
                db.get_one_metadata(conn, path=str(path))
                return True
            except ValueError:
                return False

    def get_metadata(self, path: RelativePath) -> FileMetadata:
        with get_db(self.db_path) as conn:
            metadata = db.get_one_metadata(conn, path=str(path))
            return metadata

    def _read_bytes(self, path: AbsolutePath) -> bytes:
        with open(path, "rb") as f:
            return f.read()

    def put(self, path: Path, contents: bytes) -> None:
        abs_path = self.server_settings.snapshot_folder / path
        abs_path.parent.mkdir(exist_ok=True, parents=True)

        conn = get_db(self.db_path)
        cursor = conn.cursor()
        cursor.execute("BEGIN IMMEDIATE;")
        abs_path.write_bytes(contents)
        metadata = hash_file(abs_path, root_dir=self.server_settings.snapshot_folder)
        db.save_file_metadata(cursor, metadata)
        conn.commit()
        cursor.close()
        conn.close()

    def list(self, path: RelativePath) -> list[FileMetadata]:
        with get_db(self.db_path) as conn:
            metadata = db.get_all_metadata(conn, path_like=path.as_posix())
            return metadata



from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Optional

from loguru import logger
from pydantic import BaseModel

from syftbox.client.base import SyftClientInterface
from syftbox.client.plugins.sync.endpoints import get_remote_state
from syftbox.lib.ignore import filter_ignored_paths
from syftbox.lib.lib import SyftPermission
from syftbox.server.sync.hash import hash_dir
from syftbox.server.sync.models import FileMetadata


class SyncSide(str, Enum):
    LOCAL = "local"
    REMOTE = "remote"


class FileChangeInfo(BaseModel, frozen=True):
    local_sync_folder: Path
    path: Path
    side_last_modified: SyncSide
    date_last_modified: datetime
    file_size: int = 1

    @property
    def local_abs_path(self) -> Path:
        return self.local_sync_folder / self.path

    def get_priority(self) -> int:
        if SyftPermission.is_permission_file(self.path):
            return 0
        else:
            return max(1, self.file_size)

    def __lt__(self, other: "FileChangeInfo") -> bool:
        return self.path < other.path


def format_paths(path_list: list[Path]) -> str:
    tree = ""
    folders_seen = set()

    for p in path_list:
        parts = p.parts
        for i in range(len(parts)):
            current_path = "/".join(parts[: i + 1])
            if current_path not in folders_seen:
                depth = i
                name = parts[i]
                is_file = i == len(parts) - 1
                prefix = "  " * depth + "├── "

                if is_file:
                    tree += f"{prefix}{name}\n"
                else:
                    tree += f"{prefix}{name}/\n"
                folders_seen.add(current_path)

    return tree


class DatasiteState:
    def __init__(
        self, client: SyftClientInterface, email: str, remote_state: Optional[list[FileMetadata]] = None
    ) -> None:
        """A class to represent the state of a datasite

        Args:
            ctx (SyftClientInterface): Context of the syft client
            email (str): Email of the datasite
            remote_state (Optional[list[FileMetadata]], optional): Remote state of the datasite.
                If not provided, it will be fetched from the server. Defaults to None.
        """
        self.client: SyftClientInterface = client
        self.email: str = email
        self.remote_state: Optional[list[FileMetadata]] = remote_state

    def __repr__(self) -> str:
        return f"DatasiteState<{self.email}>"

    def tree_repr(self) -> str:
        remote_state = self.remote_state or []
        rel_paths = sorted([file.path for file in remote_state])
        path_str = format_paths(rel_paths)
        return f"""DatasiteState:
{path_str}
        """

    @property
    def path(self) -> Path:
        p = self.client.workspace.datasites / self.email
        return p.expanduser().resolve()

    def get_current_local_state(self) -> list[FileMetadata]:
        return hash_dir(self.path, root_dir=self.client.workspace.datasites)

    def get_remote_state(self) -> list[FileMetadata]:
        if self.remote_state is None:
            self.remote_state = get_remote_state(self.client.server_client, path=Path(self.email))
        return self.remote_state

    def is_in_sync(self) -> bool:
        permission_changes, file_changes = self.get_out_of_sync_files()
        return len(permission_changes) == 0 and len(file_changes) == 0

    def get_out_of_sync_files(
        self,
    ) -> tuple[list[FileChangeInfo], list[FileChangeInfo]]:
        """
        calculate the files that are out of sync

        NOTE: we are not handling local permissions here,
        they will be handled by the server and consumer
        TODO: we are not handling empty folders
        """
        try:
            local_state = self.get_current_local_state()
        except Exception:
            logger.error(f"Failed to get local state for {self.email}")
            return [], []

        try:
            remote_state = self.get_remote_state()
        except Exception:
            logger.error(f"Failed to get remote state from server {self.email}")
            return [], []

        local_state_dict = {file.path: file for file in local_state}
        remote_state_dict = {file.path: file for file in remote_state}
        all_files = set(local_state_dict.keys()) | set(remote_state_dict.keys())
        all_files_filtered = filter_ignored_paths(
            datasites_dir=self.client.workspace.datasites,
            relative_paths=list(all_files),
            ignore_hidden_files=True,
            ignore_symlinks=True,
        )

        all_changes = []

        for afile in all_files_filtered:
            local_info = local_state_dict.get(afile)
            remote_info = remote_state_dict.get(afile)

            try:
                change_info = compare_fileinfo(self.client.workspace.datasites, afile, local_info, remote_info)
            except Exception as e:
                logger.error(
                    f"Failed to compare file {afile.as_posix()}, it will be retried in the next sync. Reason: {e}"
                )
                continue

            if change_info is not None:
                all_changes.append(change_info)

        # TODO implement ignore rules
        # ignore_rules = get_ignore_rules(local_state)
        # filtered_changes = filter_ignored_changes(all_changes, ignore_rules)

        permission_changes, file_changes = split_permissions(all_changes)
        # TODO debounce changes
        # filtered_changes = filter_recent_local_changes(filtered_changes)

        return permission_changes, file_changes


def split_permissions(
    changes: list[FileChangeInfo],
) -> tuple[list[FileChangeInfo], list[FileChangeInfo]]:
    permissions = []
    files = []
    for change in changes:
        if SyftPermission.is_permission_file(change.path):
            permissions.append(change)
        else:
            files.append(change)
    return permissions, files


def compare_fileinfo(
    local_sync_folder: Path,
    path: Path,
    local_info: Optional[FileMetadata],
    remote_info: Optional[FileMetadata],
) -> Optional[FileChangeInfo]:
    if local_info is None and remote_info is None:
        return

    if local_info is None and remote_info is not None:
        # File only exists on remote
        return FileChangeInfo(
            local_sync_folder=local_sync_folder,
            path=path,
            side_last_modified=SyncSide.REMOTE,
            date_last_modified=remote_info.last_modified,
            file_size=remote_info.file_size,
        )

    if remote_info is None and local_info is not None:
        # File only exists on local
        return FileChangeInfo(
            local_sync_folder=local_sync_folder,
            path=path,
            side_last_modified=SyncSide.LOCAL,
            date_last_modified=local_info.last_modified,
            file_size=local_info.file_size,
        )

    if local_info.hash != remote_info.hash:
        # File is different on both sides
        if local_info.last_modified > remote_info.last_modified:
            date_last_modified = local_info.last_modified
            side_last_modified = SyncSide.LOCAL
            file_size = local_info.file_size
        else:
            date_last_modified = remote_info.last_modified
            side_last_modified = SyncSide.REMOTE
            file_size = remote_info.file_size

        return FileChangeInfo(
            local_sync_folder=local_sync_folder,
            path=path,
            side_last_modified=side_last_modified,
            date_last_modified=date_last_modified,
            file_size=file_size,
        )



from pathlib import Path

import httpx
from typing_extensions import Protocol

from syftbox.lib.client_config import SyftClientConfig
from syftbox.lib.workspace import SyftWorkspace


class SyftClientInterface(Protocol):
    """
    Protocol defining the essential attributes required by SyftClient plugins/components.

    This interface serves two main purposes:
    1. Prevents circular dependencies by providing a minimal interface that
       plugins/components can import and type hint against, instead of importing
       the full SyftClient class.
    2. Enables dependency injection by defining a contract that any context
       or mock implementation can fulfill for testing or modular configuration.

    Attributes:
        config: Configuration settings for the Syft client
        workspace: Workspace instance managing data and computation
        server_client: HTTP client for server communication
    """

    config: SyftClientConfig
    """Configuration settings for the Syft client."""

    workspace: SyftWorkspace
    """Workspace instance managing data and computation."""

    server_client: httpx.Client

    @property
    def email(self) -> str:
        """Email address of the current user."""
        ...

    @property
    def datasite(self) -> Path:
        """Path to the datasite directory for the current user."""
        ...  # pragma: no cover

    @property
    def all_datasites(self) -> list[str]:
        """Path to the datasite directory for the current user."""
        ...  # pragma: no cover

    def log_analytics_event(self, event_name: str, **kwargs) -> None:
        """Log an analytics event to the server."""
        ...  # pragma: no cover



import enum
import hashlib
import threading
import zipfile
from enum import Enum
from io import BytesIO
from pathlib import Path
from typing import Optional

import py_fast_rsync
from loguru import logger
from pydantic import BaseModel

from syftbox.client.base import SyftClientInterface
from syftbox.client.exceptions import SyftServerError
from syftbox.client.plugins.sync.constants import MAX_FILE_SIZE_MB
from syftbox.client.plugins.sync.endpoints import (
    apply_diff,
    create,
    delete,
    download,
    download_bulk,
    get_diff,
    get_metadata,
)
from syftbox.client.plugins.sync.exceptions import FatalSyncError, SyncEnvironmentError
from syftbox.client.plugins.sync.queue import SyncQueue, SyncQueueItem
from syftbox.client.plugins.sync.sync import DatasiteState, SyncSide
from syftbox.lib.ignore import filter_ignored_paths
from syftbox.lib.lib import SyftPermission
from syftbox.server.sync.hash import hash_file
from syftbox.server.sync.models import FileMetadata


class SyncDecisionType(Enum):
    NOOP = 0
    CREATE = 1
    MODIFY = 2
    DELETE = 3


def update_local(client: SyftClientInterface, local_syncstate: FileMetadata, remote_syncstate: FileMetadata):
    diff = get_diff(client.server_client, local_syncstate.path, local_syncstate.signature_bytes)
    abs_path = client.workspace.datasites / local_syncstate.path
    local_data = abs_path.read_bytes()

    new_data = py_fast_rsync.apply(local_data, diff.diff_bytes)
    new_hash = hashlib.sha256(new_data).hexdigest()

    if new_hash != diff.hash:
        # TODO handle
        raise ValueError("hash mismatch")

    # TODO implement safe write with tempfile + rename
    abs_path.parent.mkdir(parents=True, exist_ok=True)
    abs_path.write_bytes(new_data)


def update_remote(client: SyftClientInterface, local_syncstate: FileMetadata, remote_syncstate: FileMetadata):
    abs_path = client.workspace.datasites / local_syncstate.path
    local_data = abs_path.read_bytes()

    diff = py_fast_rsync.diff(remote_syncstate.signature_bytes, local_data)
    apply_diff(client.server_client, local_syncstate.path, diff, local_syncstate.hash)


def delete_local(client: SyftClientInterface, remote_syncstate: FileMetadata):
    abs_path = client.workspace.datasites / remote_syncstate.path
    abs_path.unlink()


def delete_remote(client: SyftClientInterface, local_syncstate: FileMetadata):
    delete(client.server_client, local_syncstate.path)


def create_local(client: SyftClientInterface, remote_syncstate: FileMetadata):
    abs_path = client.workspace.datasites / remote_syncstate.path
    content_bytes = download(client.server_client, remote_syncstate.path)
    abs_path.parent.mkdir(parents=True, exist_ok=True)
    abs_path.write_bytes(content_bytes)


def create_local_batch(client: SyftClientInterface, remote_syncstates: list[Path]) -> list[str]:
    paths = [str(path) for path in remote_syncstates]
    try:
        content_bytes = download_bulk(client.server_client, paths)
    except SyftServerError as e:
        logger.error(e)
        return []
    zip_file = zipfile.ZipFile(BytesIO(content_bytes))
    zip_file.extractall(client.workspace.datasites)
    return zip_file.namelist()


def create_remote(client: SyftClientInterface, local_syncstate: FileMetadata):
    abs_path = client.workspace.datasites / local_syncstate.path
    data = abs_path.read_bytes()
    create(client.server_client, local_syncstate.path, data)


class SyncActionType(Enum):
    NOOP = enum.auto()
    CREATE_REMOTE = enum.auto()
    CREATE_LOCAL = enum.auto()
    DELETE_REMOTE = enum.auto()
    DELETE_LOCAL = enum.auto()
    MODIFY_REMOTE = enum.auto()
    MODIFY_LOCAL = enum.auto()


class SyncDecision(BaseModel):
    operation: SyncDecisionType
    side_to_update: SyncSide
    local_syncstate: Optional[FileMetadata]
    remote_syncstate: Optional[FileMetadata]
    is_executed: bool = False

    def execute(self, client: SyftClientInterface):
        if self.operation == SyncDecisionType.NOOP:
            pass
        elif self.action_type == SyncActionType.CREATE_REMOTE:
            create_remote(client, self.local_syncstate)
        elif self.action_type == SyncActionType.CREATE_LOCAL:
            create_local(client, self.remote_syncstate)
        elif self.action_type == SyncActionType.DELETE_REMOTE:
            delete_remote(client, self.remote_syncstate)
        elif self.action_type == SyncActionType.DELETE_LOCAL:
            delete_local(client, self.local_syncstate)
        elif self.action_type == SyncActionType.MODIFY_REMOTE:
            update_remote(client, self.local_syncstate, self.remote_syncstate)
        elif self.action_type == SyncActionType.MODIFY_LOCAL:
            update_local(client, self.local_syncstate, self.remote_syncstate)

        self.is_executed = True

    @property
    def path(self) -> Path:
        if self.local_syncstate:
            return self.local_syncstate.path
        elif self.remote_syncstate:
            return self.remote_syncstate.path

        raise ValueError("No path found in SyncDecision")

    @property
    def action_type(self):
        if self.operation == SyncDecisionType.NOOP:
            return SyncActionType.NOOP
        if self.operation == SyncDecisionType.CREATE and self.side_to_update == SyncSide.LOCAL:
            return SyncActionType.CREATE_LOCAL
        elif self.operation == SyncDecisionType.CREATE and self.side_to_update == SyncSide.REMOTE:
            return SyncActionType.CREATE_REMOTE
        elif self.operation == SyncDecisionType.DELETE and self.side_to_update == SyncSide.LOCAL:
            return SyncActionType.DELETE_LOCAL
        elif self.operation == SyncDecisionType.DELETE and self.side_to_update == SyncSide.REMOTE:
            return SyncActionType.DELETE_REMOTE
        elif self.operation == SyncDecisionType.MODIFY and self.side_to_update == SyncSide.LOCAL:
            return SyncActionType.MODIFY_LOCAL
        elif self.operation == SyncDecisionType.MODIFY and self.side_to_update == SyncSide.REMOTE:
            return SyncActionType.MODIFY_REMOTE

    @classmethod
    def noop(
        cls,
        local_syncstate: FileMetadata,
        remote_syncstate: FileMetadata,
    ):
        return cls(
            operation=SyncDecisionType.NOOP,
            side_to_update=SyncSide.LOCAL,
            local_syncstate=local_syncstate,
            remote_syncstate=remote_syncstate,
        )

    @classmethod
    def from_modified_states(
        cls,
        local_syncstate: Optional[FileMetadata],
        remote_syncstate: Optional[FileMetadata],
        side_to_update: SyncSide,
    ):
        """Asssumes at least on of the states is modified"""

        delete = (
            side_to_update == SyncSide.REMOTE
            and local_syncstate is None
            or side_to_update == SyncSide.LOCAL
            and remote_syncstate is None
        )

        create = (
            side_to_update == SyncSide.REMOTE
            and remote_syncstate is None
            or side_to_update == SyncSide.LOCAL
            and local_syncstate is None
        )

        if delete:
            operation = SyncDecisionType.DELETE
        elif create:
            operation = SyncDecisionType.CREATE
        else:
            operation = SyncDecisionType.MODIFY

        return cls(
            operation=operation,
            side_to_update=side_to_update,
            local_syncstate=local_syncstate,
            remote_syncstate=remote_syncstate,
        )

    def _is_invalid_remote_permission_change(self, local_abs_path: Path) -> bool:
        # we want to make sure that
        # 1) We never upload invalid syftperm files
        # 2) We allow for modifications/deletions of syftperm files, even if the local version
        # is corrupted

        if self.side_to_update != SyncSide.REMOTE:
            # Decision does not update remote, no need to check for invalid perm file
            return False

        remote_op = self.operation
        is_invalid_permission_change = (
            remote_op in [SyncDecisionType.CREATE, SyncDecisionType.MODIFY]
            and SyftPermission.is_permission_file(local_abs_path)
            and not SyftPermission.is_valid(local_abs_path)
        )
        return is_invalid_permission_change

    def _is_valid_remote_decision(self, abs_path: Path) -> tuple[bool, str]:
        if self.operation in [SyncDecisionType.NOOP, SyncDecisionType.DELETE]:
            return True, ""

        # Create/modify without file data
        if self.local_syncstate is None:
            return False, f"Attempted to sync file {abs_path} to remote, but local file data is missing."

        # Create/modify invalid permission file
        if self._is_invalid_remote_permission_change(abs_path):
            return False, f"Found invalid permission {abs_path}, permission will not be synced to remote."

        # Create/modify file over max size
        max_size_bytes = MAX_FILE_SIZE_MB * 1024 * 1024
        if self.local_syncstate.file_size > max_size_bytes:
            return False, f"File {abs_path} is larger than {MAX_FILE_SIZE_MB}MB, it will not be synced to remote."

        return True, ""

    def _is_valid_local_decision(self, abs_path: Path) -> tuple[bool, str]:
        if self.operation in [SyncDecisionType.NOOP, SyncDecisionType.DELETE]:
            return True, ""

        # Create/modify without file data
        if self.remote_syncstate is None:
            return False, f"Attempted to sync file {abs_path} to local, but remote file data is missing."

        # Create/modify file over max size
        max_size_bytes = MAX_FILE_SIZE_MB * 1024 * 1024
        if self.remote_syncstate.file_size > max_size_bytes:
            return False, f"File {abs_path} is larger than {MAX_FILE_SIZE_MB}MB, it will not be synced to local."

        return True, ""

    def is_valid(self, abs_path: Path, show_warnings: bool = False) -> bool:
        """
        Returns True if the sync decision is valid and should be executed.
        If show_warnings is True, it will log warnings for invalid decisions.

        Args:
            abs_path (Path): Absolute path of the file to sync.
            show_warnings (bool, optional): If True, a warning will be logged for invalid decisions. Defaults to False.

        Returns:
            bool: True if the decision should be executed.
        """
        if self.side_to_update == SyncSide.REMOTE:
            is_valid, reason = self._is_valid_remote_decision(abs_path)
        elif self.side_to_update == SyncSide.LOCAL:
            is_valid, reason = self._is_valid_local_decision(abs_path)
        else:
            is_valid, reason = True, ""

        if not is_valid and show_warnings:
            logger.warning(reason)

        return is_valid


class SyncDecisionTuple(BaseModel):
    remote_decision: SyncDecision
    local_decision: SyncDecision

    @property
    def result_local_state(self) -> FileMetadata:
        if self.local_decision.operation == SyncDecisionType.NOOP:
            return self.local_decision.local_syncstate
        else:
            return self.local_decision.remote_syncstate

    @property
    def is_executed(self) -> bool:
        return self.local_decision.is_executed and self.remote_decision.is_executed

    @classmethod
    def from_states(
        cls,
        current_local_syncstate: Optional[FileMetadata],
        previous_local_syncstate: Optional[FileMetadata],
        current_remote_syncstate: Optional[FileMetadata],
    ):
        def noop() -> SyncDecision:
            return SyncDecision.noop(
                local_syncstate=current_local_syncstate,
                remote_syncstate=current_remote_syncstate,
            )

        local_modified = current_local_syncstate != previous_local_syncstate
        remote_modified = previous_local_syncstate != current_remote_syncstate
        in_sync = current_remote_syncstate == current_local_syncstate
        conflict = local_modified and remote_modified and not in_sync

        path = current_local_syncstate.path if current_local_syncstate else current_remote_syncstate.path
        logger.debug(
            f"{path} local_modified: {local_modified}, remote_modified: {remote_modified}, in_sync: {in_sync}, conflict: {conflict}"
        )

        if in_sync:
            return cls(
                remote_decision=noop(),
                local_decision=noop(),
            )
        elif conflict:
            # in case of conflict we always use the server state, because it was updated earlier
            remote_decision = noop()
            # we apply the server state locally
            local_decision = SyncDecision.from_modified_states(
                local_syncstate=current_local_syncstate,
                remote_syncstate=current_remote_syncstate,
                side_to_update=SyncSide.LOCAL,
            )
            return cls(remote_decision=remote_decision, local_decision=local_decision)
        else:
            # here we can assume only one party changed
            # assert (local_modified and not server_modified) or (server_modified and not local_modified)
            if local_modified:
                return cls(
                    local_decision=noop(),
                    remote_decision=SyncDecision.from_modified_states(
                        local_syncstate=current_local_syncstate,
                        remote_syncstate=current_remote_syncstate,
                        side_to_update=SyncSide.REMOTE,
                    ),
                )
            else:
                return cls(
                    local_decision=SyncDecision.from_modified_states(
                        local_syncstate=current_local_syncstate,
                        remote_syncstate=current_remote_syncstate,
                        side_to_update=SyncSide.LOCAL,
                    ),
                    remote_decision=noop(),
                )

    def is_noop(self) -> bool:
        return (
            self.local_decision.operation == SyncDecisionType.NOOP
            and self.remote_decision.operation == SyncDecisionType.NOOP
        )

    @property
    def info_message(self) -> Optional[str]:
        messages = []
        if self.local_decision.operation != SyncDecisionType.NOOP:
            messages.append(f"Syncing {self.local_decision.path} with decision: {self.local_decision.action_type.name}")
        if self.remote_decision.operation != SyncDecisionType.NOOP:
            messages.append(
                f"Syncing {self.remote_decision.path} with decision: {self.remote_decision.action_type.name}"
            )

        return ". ".join(messages) if messages else "Syncing {self.local_decision.path} with decision: NOOP"


class LocalState(BaseModel):
    path: Path
    states: dict[Path, FileMetadata] = {}

    def insert(self, path: Path, state: FileMetadata):
        if not isinstance(path, Path):
            raise ValueError(f"path must be a Path object, got {path}")
        if not self.path.is_file():
            # If the LocalState file does not exist, the sync environment is corrupted and syncing should be aborted

            # NOTE: this can occur when the user deletes the sync folder, but a different plugin re-creates it.
            # If the sync folder exists but the LocalState file does not, it means the sync folder was deleted
            # during syncing and might cause unexpected behavior like deleting files on the remote
            raise SyncEnvironmentError("Your previous sync state has been deleted by a different process.")

        if state is None:
            self.states.pop(path, None)
        else:
            self.states[path] = state
        self.save()

    def save(self):
        try:
            with threading.Lock():
                self.path.write_text(self.model_dump_json())
        except Exception:
            logger.exception(f"Failed to save {self.path}")

    def load(self):
        with threading.Lock():
            if self.path.exists():
                data = self.path.read_text()
                loaded_state = self.model_validate_json(data)
                self.states = loaded_state.states
            else:
                self.save()


class SyncConsumer:
    def __init__(self, client: SyftClientInterface, queue: SyncQueue):
        self.client = client
        self.queue = queue
        self.previous_state = LocalState(path=Path(client.workspace.plugins) / "local_syncstate.json")
        try:
            self.previous_state.load()
        except Exception as e:
            raise SyncEnvironmentError(f"Failed to load previous sync state: {e}")

    def validate_sync_environment(self):
        if not Path(self.client.workspace.datasites).is_dir():
            raise SyncEnvironmentError("Your sync folder has been deleted by a different process.")
        if not self.previous_state.path.is_file():
            raise SyncEnvironmentError("Your previous sync state has been deleted by a different process.")

    def consume_all(self):
        while not self.queue.empty():
            self.validate_sync_environment()
            item = self.queue.get(timeout=0.1)
            try:
                self.process_filechange(item)
            except FatalSyncError as e:
                # Fatal error, syncing should be interrupted
                raise e
            except Exception as e:
                logger.error(f"Failed to sync file {item.data.path}, it will be retried in the next sync. Reason: {e}")

    def download_all_missing(self, datasite_states: list[DatasiteState]):
        try:
            missing_files: list[Path] = []
            for datasite_state in datasite_states:
                for file in datasite_state.remote_state:
                    path = file.path
                    if not self.previous_state.states.get(path):
                        missing_files.append(path)
            missing_files = filter_ignored_paths(self.client.workspace.datasites, missing_files)

            logger.info(f"Downloading {len(missing_files)} files in batch")
            received_files = create_local_batch(self.client, missing_files)
            for path in received_files:
                path = Path(path)
                state = self.get_current_local_syncstate(path)
                self.previous_state.insert(
                    path=path,
                    state=state,
                )
        except FatalSyncError as e:
            raise e
        except Exception as e:
            logger.error(
                f"Failed to download missing files, files will be downloaded individually instead. Reason: {e}"
            )

    def get_decisions(self, item: SyncQueueItem) -> SyncDecisionTuple:
        path = item.data.path
        current_local_syncstate: FileMetadata = self.get_current_local_syncstate(path)
        previous_local_syncstate = self.get_previous_local_syncstate(path)
        # TODO, rename to remote
        current_server_state = self.get_current_server_state(path)

        return SyncDecisionTuple.from_states(current_local_syncstate, previous_local_syncstate, current_server_state)

    def process_decision(self, item: SyncQueueItem, decision: SyncDecisionTuple) -> None:
        abs_path = item.data.local_abs_path

        # Ensure no changes are made if the sync environment is corrupted
        self.validate_sync_environment()
        if decision.local_decision.is_valid(abs_path=abs_path, show_warnings=True):
            decision.local_decision.execute(self.client)

        if decision.remote_decision.is_valid(abs_path=abs_path, show_warnings=True):
            decision.remote_decision.execute(self.client)

        if decision.is_executed:
            self.previous_state.insert(path=item.data.path, state=decision.result_local_state)

    def process_filechange(self, item: SyncQueueItem) -> None:
        decisions = self.get_decisions(item)
        if not decisions.is_noop():
            logger.info(decisions.info_message)
        self.process_decision(item, decisions)

    def get_current_local_syncstate(self, path: Path) -> Optional[FileMetadata]:
        abs_path = self.client.workspace.datasites / path
        if not abs_path.is_file():
            return None
        return hash_file(abs_path, root_dir=self.client.workspace.datasites)

    def get_previous_local_syncstate(self, path: Path) -> Optional[FileMetadata]:
        return self.previous_state.states.get(path, None)

    def get_current_server_state(self, path: Path) -> Optional[FileMetadata]:
        try:
            return get_metadata(self.client.server_client, path)
        except SyftServerError:
            return None



# Example test using the framework
import json
import random

import pytest
from loguru import logger

from tests.e2e.conftest import Client, E2EContext, Server


def deployment_config():
    return {
        "e2e_name": "basic_aggregator",
        "server": Server(port=5001),
        "clients": [
            Client(name="agg", port=8080, apps=["https://github.com/OpenMined/basic_aggregator"]),
            Client(name="user1", port=8081, server_port=5001, apps=["https://github.com/OpenMined/adder"]),
            Client(name="user2", port=8082, server_port=5001, apps=["https://github.com/OpenMined/adder"]),
        ],
    }


@pytest.mark.asyncio
@pytest.mark.parametrize("e2e_context", [deployment_config()], indirect=True, ids=["basic_aggregator"])
async def test_e2e_basic_aggregator(e2e_context: E2EContext):
    logger.info(f"Starting E2E '{e2e_context.e2e_name}'")
    e2e_context.reset_test_dir()

    await e2e_context.start_all()
    await e2e_context.wait_for_api("basic_aggregator", e2e_context.clients[0])

    total = 0

    # write a file "values.txt" in the public directory of both users
    logger.info("Copying values in public directories")
    for client in e2e_context.clients[1:]:
        values_txt = client.public_dir / "value.txt"
        val = random.random() * 100
        values_txt.write_text(str(val))
        total += val
        logger.debug(f"{client.name} value: {val}")

    # wait for the aggregator to finish
    logger.info("Waiting for aggregator to generate results")
    result_path = e2e_context.clients[0].api_data_dir("basic_aggregation") / "results.json"
    await e2e_context.wait_for_path(result_path, timeout=60, interval=1)

    # check the result
    result = json.loads(result_path.read_text())
    logger.info(f"Validating results\n{result}")
    assert total == result["total"]
    assert "agg@openmined.org" in result["missing"]
    assert "user1@openmined.org" in result["participants"]
    assert "user2@openmined.org" in result["participants"]



import httpx
from fastapi import APIRouter, Depends
from loguru import logger

from syftbox.server.emails.models import SendEmailRequest
from syftbox.server.settings import ServerSettings, get_server_settings

from .constants import EMAIL_SERVICE_API_URL

router = APIRouter(prefix="/emails", tags=["email"])

# TODO add some safety mechanisms to the below endpoints (rate limiting, authorization, etc)


@router.post("/")
async def send_email(
    email_request: SendEmailRequest,
    server_settings: ServerSettings = Depends(get_server_settings),
) -> bool:
    if not server_settings.email_service_api_key:
        raise httpx.HTTPStatusError("Email service API key is not set", request=None, response=None)

    async with httpx.AsyncClient() as client:
        response = await client.post(
            EMAIL_SERVICE_API_URL,
            headers={
                "Authorization": f"Bearer {server_settings.email_service_api_key}",
                "Content-Type": "application/json",
            },
            json=email_request.json_for_request(),
        )
        if response.status_code == 200:
            sent_to = email_request.to if isinstance(email_request.to, str) else ", ".join(email_request.to)
            logger.info(f"Email sent successfully to {sent_to}")
            return True
        else:
            logger.error(f"Failed to send email: {response.text}")
            return False



import json

import pytest
from fastapi.testclient import TestClient

from syftbox.server.server import app
from syftbox.server.settings import ServerSettings

TEST_DATASITE_NAME = "test_datasite@openmined.org"
TEST_FILE = "test_file.txt"
PERMFILE_FILE = "_.syftperm"
PERMFILE_DICT = {
    "admin": [TEST_DATASITE_NAME],
    "read": [TEST_DATASITE_NAME],
    "write": [TEST_DATASITE_NAME],
}


def get_access_token(client: TestClient, email: str) -> str:
    response = client.post("/auth/request_email_token", json={"email": email})
    email_token = response.json()["email_token"]
    response = client.post(
        "/auth/validate_email_token",
        headers={"Authorization": f"Bearer {email_token}"},
        params={"email": email},
    )
    if response.status_code != 200:
        raise ValueError(f"Failed to get access token, {response.text}")
    return response.json()["access_token"]


@pytest.fixture(scope="function")
def client(monkeypatch, tmp_path):
    """Every client gets their own snapshot folder at `tmp_path`"""
    snapshot_folder = tmp_path / "snapshot"
    settings = ServerSettings.from_data_folder(snapshot_folder)
    monkeypatch.setenv("SYFTBOX_DATA_FOLDER", str(settings.data_folder))
    monkeypatch.setenv("SYFTBOX_SNAPSHOT_FOLDER", str(settings.snapshot_folder))
    monkeypatch.setenv("SYFTBOX_USER_FILE_PATH", str(settings.user_file_path))

    datasite_name = TEST_DATASITE_NAME
    datasite = settings.snapshot_folder / datasite_name
    datasite.mkdir(parents=True)

    datafile = datasite / TEST_FILE
    datafile.touch()
    datafile.write_bytes(b"Hello, World!")

    datafile = datasite / TEST_DATASITE_NAME / TEST_FILE
    datafile.parent.mkdir(parents=True)

    datafile.touch()
    datafile.write_bytes(b"Hello, World!")

    permfile = datasite / PERMFILE_FILE
    permfile.touch()
    permfile.write_text(json.dumps(PERMFILE_DICT))

    with TestClient(app) as client:
        access_token = get_access_token(client, TEST_DATASITE_NAME)
        client.headers["Authorization"] = f"Bearer {access_token}"
        yield client


@pytest.fixture(scope="function")
def client_without_perms(monkeypatch, tmp_path):
    """Every client gets their own snapshot folder at `tmp_path`"""
    settings = ServerSettings.from_data_folder(tmp_path)
    monkeypatch.setenv("SYFTBOX_DATA_FOLDER", str(settings.data_folder))
    monkeypatch.setenv("SYFTBOX_SNAPSHOT_FOLDER", str(settings.snapshot_folder))
    monkeypatch.setenv("SYFTBOX_USER_FILE_PATH", str(settings.user_file_path))

    datasite_name = TEST_DATASITE_NAME
    datasite = settings.snapshot_folder / datasite_name
    datasite.mkdir(parents=True)

    datafile = datasite / TEST_FILE
    datafile.touch()
    datafile.write_bytes(b"Hello, World!")

    permfile = datasite / PERMFILE_FILE
    permfile.touch()
    permfile.write_text("")

    with TestClient(app) as client:
        access_token = get_access_token(client, TEST_DATASITE_NAME)
        client.headers["Authorization"] = f"Bearer {access_token}"
        yield client



import base64
from pathlib import Path
from typing import Any

import httpx

from syftbox.client.exceptions import SyftAuthenticationError, SyftNotFound, SyftServerError
from syftbox.server.sync.models import ApplyDiffResponse, DiffResponse, FileMetadata


def handle_json_response(endpoint: str, response: httpx.Response) -> Any:
    # endpoint only needed for error message
    if response.status_code == 200:
        return response.json()

    raise SyftServerError(f"[{endpoint}] call failed: {response.text}")


def get_access_token(client: httpx.Client, email: str) -> str:
    """Only for development purposes, should not be used in production"""
    response = client.post("/auth/request_email_token", json={"email": email})
    email_token = response.json()["email_token"]
    response = client.post("/auth/validate_email_token", headers={"Authorization": f"Bearer {email_token}"})
    return response.json()["access_token"]


def get_datasite_states(client: httpx.Client, email: str) -> dict[str, list[FileMetadata]]:
    response = client.post(
        "/sync/datasite_states",
    )

    data = handle_json_response("/sync/datasite_states", response)

    return {email: [FileMetadata(**item) for item in metadata_list] for email, metadata_list in data.items()}


def get_remote_state(client: httpx.Client, path: Path) -> list[FileMetadata]:
    response = client.post(
        "/sync/dir_state",
        params={
            "dir": str(path),
        },
    )

    response_data = handle_json_response("/dir_state", response)
    metadata_list = [FileMetadata(**item) for item in response_data]
    for item in metadata_list:
        if not hasattr(client, "metadata_cache"):
            client.metadata_cache = {}
        client.metadata_cache[item.path] = item
    return metadata_list


def get_metadata(client: httpx.Client, path: Path) -> FileMetadata:
    if hasattr(client, "metadata_cache") and path in client.metadata_cache:
        return client.metadata_cache[path]
    response = client.post(
        "/sync/get_metadata",
        json={
            "path_like": path.as_posix(),
        },
    )

    response_data = handle_json_response("/sync/get_metadata", response)

    return FileMetadata(**response_data)


def get_diff(client: httpx.Client, path: Path, signature: bytes) -> DiffResponse:
    response = client.post(
        "/sync/get_diff",
        json={
            "path": str(path),
            "signature": base64.b85encode(signature).decode("utf-8"),
        },
    )

    response_data = handle_json_response("/sync/get_diff", response)
    return DiffResponse(**response_data)


def apply_diff(client: httpx.Client, path: Path, diff: bytes, expected_hash: str) -> ApplyDiffResponse:
    response = client.post(
        "/sync/apply_diff",
        json={
            "path": str(path),
            "diff": base64.b85encode(diff).decode("utf-8"),
            "expected_hash": expected_hash,
        },
    )

    response_data = handle_json_response("/sync/apply_diff", response)
    return ApplyDiffResponse(**response_data)


def delete(client: httpx.Client, path: Path) -> None:
    response = client.post(
        "/sync/delete",
        json={
            "path": str(path),
        },
    )

    response.raise_for_status()


def create(client: httpx.Client, path: Path, data: bytes) -> None:
    response = client.post("/sync/create", files={"file": (str(path), data, "text/plain")})
    response = handle_json_response("/sync/create", response)
    return


def download(client: httpx.Client, path: Path) -> bytes:
    response = client.post(
        "/sync/download",
        json={
            "path": str(path),
        },
    )

    if response.status_code != 200:
        raise SyftNotFound(f"[/sync/download] not found on server: {path}, {response.text}")

    return response.content


def download_bulk(client: httpx.Client, paths: list[str]) -> bytes:
    response = client.post(
        "/sync/download_bulk",
        json={"paths": paths},
        timeout=30,
    )
    response.raise_for_status()
    return response.content


def whoami(client: httpx.Client) -> str:
    """
    Performs a health check on the server by sending a POST request to the '/auth/whoami' endpoint.

    Args:
        client (httpx.Client): Client to use for the health check.

    Raises:
        AuthError: If the server responds with a 401 status code.
        SyftServerError: If the server responds with any other non-200 status code.

    Returns:
        None: If the health check is successful.
    """
    try:
        response = client.post("/auth/whoami")
        if response.status_code == 200:
            email = response.json()["email"]
            return email
        elif response.status_code == 401:
            raise SyftAuthenticationError()
        else:
            raise SyftServerError(
                f"Health check failed. Status code: {response.status_code}. Response: {response.text}"
            )
    except httpx.RequestError as e:
        raise SyftServerError(f"Health check failed, could not connect to server. Reason: {e}")



from typing import Optional

import httpx
import typer
from rich import print as rprint
from rich.prompt import Prompt

from syftbox.lib.client_config import SyftClientConfig


def has_valid_access_token(conf: SyftClientConfig, auth_client: httpx.Client) -> bool:
    """Returns True if conf has a valid access token that matches the email in the config."""
    if not conf.access_token:
        return False
    response = auth_client.post("/auth/whoami", headers={"Authorization": f"Bearer {conf.access_token}"})
    if response.status_code == 401:
        rprint("[red]Invalid access token, re-authenticating.[/red]")
        return False
    elif response.status_code >= 400:
        rprint(f"[red]An unexpected error occurred: {response.text}, re-authenticating.[/red]")
        return False

    authed_email = response.json().get("email", None)
    is_valid = authed_email == conf.email
    if not is_valid:
        rprint(
            f"[red]Invalid access token for {conf.email}, this token is for {authed_email}. re-authenticating.[/red]"
        )
    return is_valid


def request_email_token(auth_client: httpx.Client, conf: SyftClientConfig) -> Optional[str]:
    """
    if auth is enabled, send an email token to the user's email address.
    if auth is disabled, the token will be returned directly in the response instead.

    Args:
        auth_client (httpx.Client): httpx client
        conf (SyftClientConfig): client config

    Returns:
        Optional[str]: email token if auth is disabled, None if auth is enabled
    """
    response = auth_client.post("/auth/request_email_token", json={"email": conf.email})
    response.raise_for_status()
    return response.json().get("email_token", None)


def get_access_token(
    conf: SyftClientConfig,
    auth_client: httpx.Client,
    email_token: Optional[str] = None,
) -> str:
    """
    Validate the email token and return the access token.

    Args:
        auth_client (httpx.Client): httpx client
        email_token (Optional[str]): Optional email token. If not provided,
            the user will be prompted to input it.

    Returns:
        str: access token
    """
    if not email_token:
        email_token = Prompt.ask(
            f"[yellow]Please enter the token sent to {conf.email}. Also check your spam folder[/yellow]"
        )

    response = auth_client.post(
        "/auth/validate_email_token",
        headers={"Authorization": f"Bearer {email_token}"},
        params={"email": conf.email},
    )

    if response.status_code == 200:
        return response.json()["access_token"]
    elif response.status_code == 401:
        rprint("[red]Invalid token, please copy the full token from your email[/red]")
        return get_access_token(conf, auth_client)
    else:
        rprint(f"[red]An unexpected error occurred: {response.text}[/red]")
        typer.Exit(1)


def authenticate_user(conf: SyftClientConfig) -> str:
    auth_client = httpx.Client(base_url=str(conf.server_url))

    if has_valid_access_token(conf, auth_client):
        return conf.access_token

    email_token = request_email_token(auth_client, conf)
    return get_access_token(conf, auth_client, email_token)



from __future__ import annotations

import json
import os
from dataclasses import dataclass, field
from pathlib import Path

from loguru import logger
from typing_extensions import Any, Optional, Self, Union

from syftbox.lib.constants import PERM_FILE
from syftbox.server.sync.models import FileMetadata

USER_GROUP_GLOBAL = "GLOBAL"

ICON_FILE = "Icon"  # special
IGNORE_FILES = []


def perm_file_path(path: str) -> str:
    return os.path.join(path, PERM_FILE)


def is_primitive_json_serializable(obj):
    if isinstance(obj, (str, int, float, bool, type(None))):
        return True
    return False


def pack(obj) -> Any:
    if is_primitive_json_serializable(obj):
        return obj

    if hasattr(obj, "to_dict"):
        return obj.to_dict()

    if isinstance(obj, list):
        return [pack(val) for val in obj]

    if isinstance(obj, dict):
        return {k: pack(v) for k, v in obj.items()}

    if isinstance(obj, Path):
        return str(obj)

    raise Exception(f"Unable to pack type: {type(obj)} value: {obj}")


class Jsonable:
    def to_dict(self) -> dict:
        output = {}
        for k, v in self.__dict__.items():
            if k.startswith("_"):
                continue
            output[k] = pack(v)
        return output

    def __iter__(self):
        for key, val in self.to_dict().items():
            if key.startswith("_"):
                yield key, val

    def __getitem__(self, key):
        if key.startswith("_"):
            return None
        return self.to_dict()[key]

    @classmethod
    def load(cls, file_or_bytes: Union[str, Path, bytes]) -> Self:
        try:
            if isinstance(file_or_bytes, (str, Path)):
                with open(file_or_bytes) as f:
                    data = f.read()
            else:
                data = file_or_bytes
            d = json.loads(data)
            return cls(**d)
        except Exception as e:
            raise e

    def save(self, filepath: str) -> None:
        d = self.to_dict()
        with open(Path(filepath).expanduser(), "w") as f:
            f.write(json.dumps(d))


@dataclass
class SyftPermission(Jsonable):
    admin: list[str]
    read: list[str]
    write: list[str]
    filepath: Optional[str] = None
    terminal: bool = False

    @classmethod
    def datasite_default(cls, email: str) -> Self:
        return cls(
            admin=[email],
            read=[email],
            write=[email],
        )

    @staticmethod
    def is_permission_file(path: Union[Path, str], check_exists: bool = False) -> bool:
        path = Path(path)
        if check_exists and not path.is_file():
            return False
        return path.name == "_.syftperm"

    @classmethod
    def is_valid(cls, path_or_bytes: Union[str, Path, bytes]):
        try:
            SyftPermission.load(path_or_bytes)
            return True
        except Exception:
            return False

    def is_admin(self, email: str) -> bool:
        return email in self.admin

    def has_read_permission(self, email: str) -> bool:
        return email in self.read or USER_GROUP_GLOBAL in self.read

    def has_write_permission(self, email: str) -> bool:
        return email in self.write or USER_GROUP_GLOBAL in self.write

    def __eq__(self, other):
        if not isinstance(other, SyftPermission):
            return NotImplemented
        return (
            self.admin == other.admin
            and self.read == other.read
            and self.write == other.write
            and self.filepath == other.filepath
            and self.terminal == other.terminal
        )

    def perm_path(self, path=None) -> str:
        if path is not None:
            self.filepath = path

        if self.filepath is None:
            raise Exception(f"Saving requites a path: {self}")

        if os.path.isdir(self.filepath):
            self.filepath = perm_file_path(self.filepath)
        return self.filepath

    def save(self, path=None) -> bool:
        self.perm_path(path=path)
        if str(self.filepath).endswith(".syftperm"):
            super().save(self.filepath)
        else:
            raise Exception(f"Perm file must end in .syftperm. {self.filepath}")
        return True

    def ensure(self, path=None) -> bool:
        # make sure the contents matches otherwise write it
        self.perm_path(path=path)
        try:
            prev_perm_file = SyftPermission.load(self.filepath)
            if self == prev_perm_file:
                # no need to write
                return True
        except Exception:
            pass
        return self.save(path)

    @classmethod
    def no_permission(cls) -> Self:
        return cls(admin=[], read=[], write=[])

    @classmethod
    def mine_no_permission(cls, email: str) -> Self:
        return cls(admin=[email], read=[], write=[])

    @classmethod
    def mine_with_public_read(cls, email: str) -> Self:
        return cls(admin=[email], read=[email, "GLOBAL"], write=[email])

    @classmethod
    def mine_with_public_write(cls, email: str) -> Self:
        return cls(admin=[email], read=[email, "GLOBAL"], write=[email, "GLOBAL"])

    @classmethod
    def theirs_with_my_read(cls, their_email, my_email: str) -> Self:
        return cls(admin=[their_email], read=[their_email, my_email], write=[their_email])

    @classmethod
    def theirs_with_my_read_write(cls, their_email, my_email: str) -> Self:
        return cls(
            admin=[their_email],
            read=[their_email, my_email],
            write=[their_email, my_email],
        )

    def __repr__(self) -> str:
        string = "SyftPermission:\n"
        string += f"{self.filepath}\n"
        string += "ADMIN: ["
        for v in self.admin:
            string += v + ", "
        string += "]\n"

        string += "READ: ["
        for r in self.read:
            string += r + ", "
        string += "]\n"

        string += "WRITE: ["
        for w in self.write:
            string += w + ", "
        string += "]\n"
        return string


def get_datasites(sync_folder: Union[str, Path]) -> list[str]:
    sync_folder = str(sync_folder.resolve()) if isinstance(sync_folder, Path) else sync_folder
    datasites = []
    folders = os.listdir(sync_folder)
    for folder in folders:
        if "@" in folder:
            datasites.append(folder)
    return datasites


def build_tree_string(paths_dict, prefix=""):
    lines = []
    items = list(paths_dict.items())

    for index, (key, value) in enumerate(items):
        # Determine if it's the last item in the current directory level
        connector = "└── " if index == len(items) - 1 else "├── "
        lines.append(f"{prefix}{connector}{repr(key)}")

        # Prepare the prefix for the next level
        if isinstance(value, dict):
            extension = "    " if index == len(items) - 1 else "│   "
            lines.append(build_tree_string(value, prefix + extension))

    return "\n".join(lines)


@dataclass
class PermissionTree(Jsonable):
    tree: dict[str, SyftPermission]
    parent_path: str
    root_perm: Optional[SyftPermission]

    corrupted_permission_files: list[str] = field(default_factory=list)

    @classmethod
    def from_path(cls, parent_path, raise_on_corrupted_files: bool = False) -> Self:
        corrupted_permission_files = []
        perm_dict = {}
        for root, dirs, files in os.walk(parent_path):
            for file in files:
                if file.endswith(".syftperm"):
                    path = os.path.join(root, file)
                    try:
                        perm_dict[path] = SyftPermission.load(path)
                    except Exception:
                        corrupted_permission_files.append(path)

        root_perm = None
        root_perm_path = perm_file_path(parent_path)
        if root_perm_path in perm_dict:
            root_perm = perm_dict[root_perm_path]

        if corrupted_permission_files:
            if raise_on_corrupted_files:
                raise ValueError(f"Found corrupted permission files: {corrupted_permission_files}")
            logger.warning(f"Found corrupted permission files: {corrupted_permission_files}")

        return cls(
            root_perm=root_perm,
            tree=perm_dict,
            parent_path=parent_path,
            corrupted_permission_files=corrupted_permission_files,
        )

    def has_corrupted_permission(self, path: Union[str, Path]) -> bool:
        path = Path(path).resolve()
        corrupted_permission_paths = [Path(p).parent.resolve() for p in self.corrupted_permission_files]
        for perm_path in corrupted_permission_paths:
            if path.is_relative_to(perm_path):
                return True
        return False

    @property
    def root_or_default(self) -> SyftPermission:
        if self.root_perm:
            return self.root_perm
        return SyftPermission.no_permission()

    def permission_for_path(self, path: str) -> SyftPermission:
        parent_path = os.path.normpath(self.parent_path)
        current_perm = self.root_or_default

        # default
        if parent_path not in path:
            return current_perm

        sub_path = path.replace(parent_path, "")
        current_perm_level = parent_path
        for part in sub_path.split("/"):
            if part == "":
                continue

            current_perm_level += "/" + part
            next_perm_file = perm_file_path(current_perm_level)
            if next_perm_file in self.tree:
                next_perm = self.tree[next_perm_file]
                current_perm = next_perm

            if current_perm.terminal:
                return current_perm

        return current_perm

    def __repr__(self) -> str:
        return f"PermissionTree: {self.parent_path}\n" + build_tree_string(self.tree)


def filter_metadata(
    user_email: str,
    metadata_list: list[FileMetadata],
    perm_tree: PermissionTree,
    snapshot_folder: Path,
) -> list[FileMetadata]:
    filtered_metadata = []
    for metadata in metadata_list:
        perm_file_at_path = perm_tree.permission_for_path((snapshot_folder / metadata.path).as_posix())
        if (
            user_email in perm_file_at_path.read
            or "GLOBAL" in perm_file_at_path.read
            or user_email in perm_file_at_path.admin
        ):
            filtered_metadata.append(metadata)
    return filtered_metadata



import base64
import hashlib
import zipfile
from io import BytesIO
from pathlib import Path

import py_fast_rsync
import pytest
from fastapi.testclient import TestClient
from py_fast_rsync import signature

from syftbox.client.exceptions import SyftServerError
from syftbox.client.plugins.sync.endpoints import (
    apply_diff,
    download_bulk,
    get_datasite_states,
    get_diff,
    get_metadata,
    get_remote_state,
)
from syftbox.lib.lib import FileMetadata
from syftbox.server.sync.models import ApplyDiffResponse, DiffResponse
from tests.unit.server.conftest import PERMFILE_FILE, TEST_DATASITE_NAME, TEST_FILE


def test_get_diff_2(client: TestClient):
    local_data = b"This is my local data"
    sig = signature.calculate(local_data)
    sig_b85 = base64.b85encode(sig).decode("utf-8")
    response = client.post(
        "/sync/get_diff",
        json={
            "path": f"{TEST_DATASITE_NAME}/{TEST_FILE}",
            "signature": sig_b85,
        },
    )

    response.raise_for_status()
    diff_response = DiffResponse.model_validate(response.json())
    remote_diff = diff_response.diff_bytes
    probably_remote_data = py_fast_rsync.apply(local_data, remote_diff)

    server_settings = client.app_state["server_settings"]
    file_server_contents = server_settings.read(f"{TEST_DATASITE_NAME}/{TEST_FILE}")
    assert file_server_contents == probably_remote_data


def file_digest(file_path, algorithm="sha256"):
    # because this doesnt work in python <=3.10, we implement it manually
    hash_func = hashlib.new(algorithm)

    with open(file_path, "rb") as file:
        # Read the file in chunks to handle large files efficiently
        for chunk in iter(lambda: file.read(4096), b""):
            hash_func.update(chunk)

    return hash_func.hexdigest()


def test_syft_client_push_flow(client: TestClient):
    response = client.post(
        "/sync/get_metadata",
        json={"path_like": f"{TEST_DATASITE_NAME}/{TEST_FILE}"},
    )

    response.raise_for_status()
    server_signature_b85 = response.json()["signature"]
    server_signature = base64.b85decode(server_signature_b85)
    assert server_signature

    local_data = b"This is my local data"
    delta = py_fast_rsync.diff(server_signature, local_data)
    delta_b85 = base64.b85encode(delta).decode("utf-8")
    expected_hash = hashlib.sha256(local_data).hexdigest()

    response = client.post(
        "/sync/apply_diff",
        json={
            "path": f"{TEST_DATASITE_NAME}/{TEST_FILE}",
            "diff": delta_b85,
            "expected_hash": expected_hash,
        },
    )

    response.raise_for_status()

    result = response.json()
    snapshot_folder = client.app_state["server_settings"].snapshot_folder
    sha256local = file_digest(f"{snapshot_folder}/{TEST_DATASITE_NAME}/{TEST_FILE}", "sha256")
    assert result["current_hash"] == expected_hash == sha256local


def test_get_remote_state(client: TestClient):
    metadata = get_remote_state(client, Path(TEST_DATASITE_NAME))

    assert len(metadata) == 3


def test_get_metadata(client: TestClient):
    metadata = get_metadata(client, Path(TEST_DATASITE_NAME) / TEST_FILE)
    assert metadata.path == Path(TEST_DATASITE_NAME) / TEST_FILE

    # check serde works
    assert isinstance(metadata.hash_bytes, bytes)
    assert isinstance(metadata.signature_bytes, bytes)


def test_apply_diff(client: TestClient):
    local_data = b"This is my local data"

    remote_metadata = get_metadata(client, Path(TEST_DATASITE_NAME) / TEST_FILE)

    diff = py_fast_rsync.diff(remote_metadata.signature_bytes, local_data)
    expected_hash = hashlib.sha256(local_data).hexdigest()

    # Apply local_data to server
    response = apply_diff(client, Path(TEST_DATASITE_NAME) / TEST_FILE, diff, expected_hash)
    assert response.current_hash == expected_hash

    # check file was written correctly
    snapshot_file_path = client.app_state["server_settings"].snapshot_folder / Path(TEST_DATASITE_NAME) / TEST_FILE
    remote_data = snapshot_file_path.read_bytes()
    assert local_data == remote_data

    # another diff with incorrect hash
    remote_metadata = get_metadata(client, Path(TEST_DATASITE_NAME) / TEST_FILE)
    diff = py_fast_rsync.diff(remote_metadata.signature_bytes, local_data)
    wrong_hash = "wrong_hash"

    with pytest.raises(SyftServerError):
        apply_diff(client, Path(TEST_DATASITE_NAME) / TEST_FILE, diff, wrong_hash)


def test_get_diff(client: TestClient):
    local_data = b"This is my local data"
    sig = signature.calculate(local_data)

    file_path = Path(TEST_DATASITE_NAME) / TEST_FILE
    response = get_diff(client, file_path, sig)
    assert response.path == file_path

    # apply and check hash
    new_data = py_fast_rsync.apply(local_data, base64.b85decode(response.diff))
    new_hash = hashlib.sha256(new_data).hexdigest()

    assert new_hash == response.hash

    # diff nonexistent file
    file_path = Path(TEST_DATASITE_NAME) / "nonexistent_file.txt"
    with pytest.raises(SyftServerError):
        get_diff(client, file_path, sig)


def test_delete_file(client: TestClient):
    response = client.post(
        "/sync/delete",
        json={"path": f"{TEST_DATASITE_NAME}/{TEST_FILE}"},
    )

    response.raise_for_status()
    snapshot_folder = client.app_state["server_settings"].snapshot_folder
    path = Path(f"{snapshot_folder}/{TEST_DATASITE_NAME}/{TEST_FILE}")
    assert not path.exists()

    with pytest.raises(SyftServerError):
        get_metadata(client, Path(TEST_DATASITE_NAME) / TEST_FILE)


def test_create_file(client: TestClient):
    snapshot_folder = client.app_state["server_settings"].snapshot_folder
    new_fname = "new.txt"
    contents = b"Some content"
    path = Path(f"{snapshot_folder}/{TEST_DATASITE_NAME}/{new_fname}")
    assert not path.exists()

    with open(path, "wb") as f:
        f.write(contents)

    with open(path, "rb") as f:
        files = {"file": (f"{TEST_DATASITE_NAME}/{new_fname}", f.read())}
        response = client.post("/sync/create", files=files)
    response.raise_for_status()
    assert path.exists()


def test_permfile(client: TestClient):
    invalid_contents = b"wrong permfile"
    folder = "test"

    # invalid
    files = {"file": (f"{TEST_DATASITE_NAME}/{folder}/{PERMFILE_FILE}", invalid_contents)}
    with pytest.raises(Exception):
        response = client.post("/sync/create", files=files)
        response.raise_for_status()

    # valid
    valid_contents = b'{"admin": ["x@x.org"], "read": ["x@x.org"], "write": ["x@x.org"], "filepath": "~/_.syftperm", "terminal": false}'
    files = {"file": (f"{TEST_DATASITE_NAME}/{folder}/{PERMFILE_FILE}", valid_contents)}
    response = client.post("/sync/create", files=files)
    response.raise_for_status()


def test_update_permfile_success(client: TestClient):
    local_data = b'{"admin": ["x@x.org"], "read": ["x@x.org"], "write": ["x@x.org"], "filepath": "~/_.syftperm", "terminal": false}'

    remote_metadata = get_metadata(client, Path(TEST_DATASITE_NAME) / PERMFILE_FILE)

    diff = py_fast_rsync.diff(remote_metadata.signature_bytes, local_data)
    expected_hash = hashlib.sha256(local_data).hexdigest()

    response = apply_diff(client, Path(TEST_DATASITE_NAME) / PERMFILE_FILE, diff, expected_hash)
    assert isinstance(response, ApplyDiffResponse)


def test_update_permfile_failure(client: TestClient):
    local_data = b'3gwrehtytrterfewdw ["x@x.org"], "read": ["x@x.org"], "write": ["x@x.org"], "filepath": "~/_.syftperm", "terminal": false}'

    remote_metadata = get_metadata(client, Path(TEST_DATASITE_NAME) / PERMFILE_FILE)

    diff = py_fast_rsync.diff(remote_metadata.signature_bytes, local_data)
    expected_hash = hashlib.sha256(local_data).hexdigest()

    with pytest.raises(SyftServerError):
        apply_diff(client, Path(TEST_DATASITE_NAME) / PERMFILE_FILE, diff, expected_hash)


def test_list_datasites(client: TestClient):
    response = client.post("/sync/datasites")

    response.raise_for_status()


def test_get_all_datasite_states(client: TestClient):
    response = get_datasite_states(client, email=TEST_DATASITE_NAME)
    assert len(response) == 1

    metadatas = response[TEST_DATASITE_NAME]
    assert len(metadatas) == 3
    assert all(isinstance(m, FileMetadata) for m in metadatas)


def test_download_snapshot(client: TestClient):
    metadata = get_remote_state(client, Path(TEST_DATASITE_NAME))
    paths = [m.path.as_posix() for m in metadata]
    data = download_bulk(client, paths)
    zip_file = zipfile.ZipFile(BytesIO(data))
    assert len(zip_file.filelist) == 3


def test_whoami(client: TestClient):
    response = client.post("/auth/whoami")
    response.raise_for_status()
    assert response.json() == {"email": TEST_DATASITE_NAME}



from pathlib import Path

from typer.testing import CliRunner

from syftbox.client.cli import app as client_cli

# Initialize test runner
runner = CliRunner()


def mock_port_in_use(*args, **kwargs):
    return False


def test_client_success(monkeypatch, mock_config):
    def setup_config_interactive(*args, **kwargs):
        return mock_config

    def mock_run_client(*args, **kwargs):
        return 0

    def get_migration_decision(*args, **kwargs):
        return False

    monkeypatch.setattr("syftbox.client.client2.run_client", mock_run_client)
    monkeypatch.setattr("syftbox.client.cli_setup.setup_config_interactive", setup_config_interactive)
    monkeypatch.setattr("syftbox.client.cli_setup.get_migration_decision", get_migration_decision)
    monkeypatch.setattr("syftbox.client.utils.net.is_port_in_use", mock_port_in_use)

    result = runner.invoke(client_cli)
    assert result.exit_code == 0


def test_client_error(monkeypatch, mock_config):
    def setup_config_interactive(*args, **kwargs):
        return mock_config

    def mock_run_client(*args, **kwargs):
        return -1

    def get_migration_decision(*args, **kwargs):
        return False

    monkeypatch.setattr("syftbox.client.client2.run_client", mock_run_client)
    monkeypatch.setattr("syftbox.client.cli_setup.setup_config_interactive", setup_config_interactive)
    monkeypatch.setattr("syftbox.client.utils.net.is_port_in_use", mock_port_in_use)
    monkeypatch.setattr("syftbox.client.cli_setup.get_migration_decision", get_migration_decision)

    result = runner.invoke(client_cli)
    assert result.exit_code == -1


def test_port_error(monkeypatch):
    monkeypatch.setattr("syftbox.client.utils.net.is_port_in_use", lambda p: True)
    result = runner.invoke(client_cli)
    assert result.exit_code == 1


def test_client_report(monkeypatch, tmp_path, mock_config):
    monkeypatch.setattr("syftbox.client.logger.zip_logs", lambda p, **kw: Path(str(p) + ".log"))
    result = runner.invoke(client_cli, ["report"])
    assert result.exit_code == 0



import json
from pathlib import Path


class AppMockFactory:
    @staticmethod
    def create_app_without_config(apps_dir: Path, app_name: str) -> tuple[Path, str]:
        """Create a mock app without configuration."""
        mock_app_dir = Path(apps_dir) / app_name
        script_output = "Test executed"

        mock_app_dir.mkdir(parents=True, exist_ok=True)
        run_script = mock_app_dir / "run.sh"

        run_script.write_text(f"#!/bin/bash\necho '{script_output}' | tee app.log")
        run_script.chmod(0o755)

        return mock_app_dir, script_output

    @staticmethod
    def create_app_with_config(apps_dir: Path, app_name: str) -> tuple[Path, str]:
        """Create a mock app with configuration."""
        mock_app_dir = Path(apps_dir) / app_name
        mock_app_dir.mkdir(parents=True, exist_ok=True)

        test_value = "test_value"
        config_json = {
            "app": {
                "env": {"TEST_VAR": test_value},
                "run": {"schedule": None, "interval": 1, "command": ["./test.sh"]},
            }
        }

        config_file = mock_app_dir / "config.json"
        config_file.write_text(json.dumps(config_json))

        test_script = mock_app_dir / "test.sh"
        test_script.write_text("#!/bin/bash\necho $TEST_VAR | tee app.log")
        test_script.chmod(0o755)

        return mock_app_dir, test_value



import time

from fastapi import Request
from loguru import logger
from starlette.middleware.base import BaseHTTPMiddleware


class LoguruMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request: Request, call_next):
        start_time = time.time()
        response = await call_next(request)
        duration = time.time() - start_time
        logger.info(f"{request.method} {request.url.path} {response.status_code} {duration:.2f}s")

        return response



import json
import os
import shutil
import time
from pathlib import Path

import faker
import pytest
from fastapi.testclient import TestClient
from loguru import logger

from syftbox.client.base import SyftClientInterface
from syftbox.client.plugins.sync.constants import MAX_FILE_SIZE_MB
from syftbox.client.plugins.sync.exceptions import FatalSyncError
from syftbox.client.plugins.sync.manager import DatasiteState, SyncManager, SyncQueueItem
from syftbox.client.utils.dir_tree import DirTree, create_dir_tree
from syftbox.lib.lib import SyftPermission
from syftbox.server.settings import ServerSettings
from tests.unit.sync.conftest import setup_datasite

fake = faker.Faker()


# def create_random_file(client_config: ClientConfig, sub_path: str = "") -> Path:
#     relative_path = Path(sub_path) / fake.file_name(extension="json")
#     file_path = client_config.datasite / relative_path
#     content = {"body": fake.text()}
#     file_path.write_text(json.dumps(content))

#     path_in_datasite = file_path.relative_to(client_config.workspace.datasites)
#     return path_in_datasite


def assert_files_not_on_datasite(client: SyftClientInterface, files: list[Path]):
    for file in files:
        assert not (client.workspace.datasites / file).exists(), f"File {file} exists on datasite {client.email}"


def assert_files_on_datasite(client: SyftClientInterface, files: list[Path]):
    for file in files:
        assert (client.workspace.datasites / file).exists(), f"File {file} does not exist on datasite {client.email}"


def assert_files_on_server(server_client: TestClient, files: list[Path]):
    server_settings: ServerSettings = server_client.app_state["server_settings"]
    for file in files:
        assert (server_settings.snapshot_folder / file).exists(), f"File {file} does not exist on server"


def assert_dirtree_exists(base_path: Path, tree: DirTree) -> None:
    for name, content in tree.items():
        local_path = base_path / name

        if isinstance(content, str):
            assert local_path.read_text() == content
        elif isinstance(content, SyftPermission):
            assert json.loads(local_path.read_text()) == content.to_dict()
        elif isinstance(content, dict):
            assert local_path.is_dir()
            assert_dirtree_exists(local_path, content)


def test_get_datasites(datasite_1: SyftClientInterface, datasite_2: SyftClientInterface):
    emails = {datasite_1.email, datasite_2.email}
    sync_service = SyncManager(datasite_1)
    sync_service2 = SyncManager(datasite_2)
    sync_service.run_single_thread()
    sync_service2.run_single_thread()

    datasites = sync_service.get_datasite_states()
    assert {datasites[0].email, datasites[1].email} == emails


def test_enqueue_changes(datasite_1: SyftClientInterface):
    sync_service = SyncManager(datasite_1)
    datasites = sync_service.get_datasite_states()

    out_of_sync_permissions, out_of_sync_files = datasites[0].get_out_of_sync_files()
    num_files_after_setup = len(out_of_sync_files) + len(out_of_sync_permissions)

    # Create two files in datasite_1
    tree = {
        "folder1": {
            "_.syftperm": SyftPermission.mine_with_public_read(datasite_1.email),
            "large.txt": fake.text(max_nb_chars=1000),
            "small.txt": fake.text(max_nb_chars=10),
        },
    }
    create_dir_tree(Path(datasite_1.datasite), tree)
    out_of_sync_permissions, out_of_sync_files = datasites[0].get_out_of_sync_files()
    num_out_of_sync_files = len(out_of_sync_files) + len(out_of_sync_permissions)
    # 3 new files
    assert num_out_of_sync_files - num_files_after_setup == 3

    # Enqueue the changes + verify order
    for change in out_of_sync_permissions + out_of_sync_files:
        sync_service.enqueue(change)

    items_from_queue: list[SyncQueueItem] = []
    while not sync_service.queue.empty():
        items_from_queue.append(sync_service.queue.get())

    should_be_permissions = items_from_queue[: len(out_of_sync_permissions)]
    should_be_files = items_from_queue[len(out_of_sync_permissions) :]

    assert all(SyftPermission.is_permission_file(item.data.path) for item in should_be_permissions)
    assert all(not SyftPermission.is_permission_file(item.data.path) for item in should_be_files)

    for item in should_be_files:
        print(item.priority, item.data)


def test_create_file(server_client: TestClient, datasite_1: SyftClientInterface, datasite_2: SyftClientInterface):
    server_settings: ServerSettings = server_client.app_state["server_settings"]
    sync_service = SyncManager(datasite_1)

    # Create a file in datasite_1
    tree = {
        "folder1": {
            "_.syftperm": SyftPermission.mine_with_public_read(datasite_1.email),
            "file.txt": fake.text(max_nb_chars=1000),
        },
    }
    create_dir_tree(Path(datasite_1.datasite), tree)

    # changes are pushed to server
    sync_service.run_single_thread()

    # check if no changes are left
    for datasite in sync_service.get_datasite_states():
        out_of_sync_permissions, out_of_sync_files = datasite.get_out_of_sync_files()
        assert not out_of_sync_files
        assert not out_of_sync_permissions

    # check if file exists on server
    print(datasite_2.workspace.datasites)
    datasite_snapshot = server_settings.snapshot_folder / datasite_1.email
    assert_dirtree_exists(datasite_snapshot, tree)

    # check if file exists on datasite_2
    sync_client_2 = SyncManager(datasite_2)
    sync_client_2.run_single_thread()
    datasite_states = sync_client_2.get_datasite_states()
    ds1_state = datasite_states[0]
    assert ds1_state.email == datasite_1.email

    print(ds1_state.get_out_of_sync_files())

    print(f"datasites {[d.email for d in sync_client_2.get_datasite_states()]}")
    sync_client_2.run_single_thread()

    assert_files_on_datasite(datasite_2, [Path(datasite_1.email) / "folder1" / "file.txt"])


def test_modify(server_client: TestClient, datasite_1: SyftClientInterface):
    server_settings: ServerSettings = server_client.app_state["server_settings"]
    sync_service_1 = SyncManager(datasite_1)

    # Setup initial state
    tree = {
        "folder1": {
            "_.syftperm": SyftPermission.mine_with_public_write(datasite_1.email),
            "file.txt": "content",
        },
    }
    create_dir_tree(Path(datasite_1.datasite), tree)
    sync_service_1.run_single_thread()

    # modify the file
    file_path = datasite_1.datasite / "folder1" / "file.txt"
    new_content = "modified"
    file_path.write_text(new_content)
    assert file_path.read_text() == new_content

    sync_service_1.run_single_thread()

    assert file_path.read_text() == new_content
    assert (server_settings.snapshot_folder / datasite_1.email / "folder1" / "file.txt").read_text() == new_content


def test_modify_and_pull(server_client: TestClient, datasite_1: SyftClientInterface, datasite_2: SyftClientInterface):
    server_settings: ServerSettings = server_client.app_state["server_settings"]
    sync_service_1 = SyncManager(datasite_1)
    sync_service_2 = SyncManager(datasite_2)

    # Setup initial state
    tree = {
        "folder1": {
            "_.syftperm": SyftPermission.mine_with_public_write(datasite_1.email),
            "file.txt": "content1",
        },
    }
    create_dir_tree(Path(datasite_1.datasite), tree)
    sync_service_1.run_single_thread()
    sync_service_2.run_single_thread()

    # modify the file
    file_path = datasite_1.datasite / "folder1" / "file.txt"
    new_content = fake.text(max_nb_chars=100_000)
    file_path.write_text(new_content)

    assert file_path.read_text() == new_content

    sync_service_1.run_single_thread()

    assert file_path.read_text() == new_content
    assert (server_settings.snapshot_folder / datasite_1.email / "folder1" / "file.txt").read_text() == new_content

    sync_service_2.run_single_thread()

    assert file_path.read_text() == new_content
    assert (Path(datasite_2.workspace.datasites) / datasite_1.email / "folder1" / "file.txt").read_text() == new_content


def test_modify_with_conflict(
    server_client: TestClient, datasite_1: SyftClientInterface, datasite_2: SyftClientInterface
):
    sync_service_1 = SyncManager(datasite_1)
    sync_service_2 = SyncManager(datasite_2)

    # Setup initial state
    tree = {
        "folder1": {
            "_.syftperm": SyftPermission.mine_with_public_write(datasite_1.email),
            "file.txt": "content1",
        },
    }
    create_dir_tree(Path(datasite_1.datasite), tree)
    sync_service_1.run_single_thread()
    sync_service_2.run_single_thread()

    # modify the file both clients
    file_path_1 = datasite_1.datasite / "folder1" / "file.txt"
    new_content_1 = "modified1"
    file_path_1.write_text(new_content_1)

    file_path_2 = Path(datasite_2.workspace.datasites) / datasite_1.email / "folder1" / "file.txt"
    new_content_2 = "modified2"
    file_path_2.write_text(new_content_2)

    assert new_content_1 != new_content_2
    assert file_path_1.read_text() == new_content_1
    assert file_path_2.read_text() == new_content_2

    # first to server wins
    sync_service_1.run_single_thread()
    sync_service_2.run_single_thread()

    assert file_path_1.read_text() == new_content_1
    assert file_path_2.read_text() == new_content_1

    # modify again, 2 syncs first
    new_content_1 = fake.text(max_nb_chars=1000)
    new_content_2 = fake.text(max_nb_chars=1000)
    file_path_1.write_text(new_content_1)
    file_path_2.write_text(new_content_2)
    assert new_content_1 != new_content_2

    assert file_path_1.read_text() == new_content_1
    assert file_path_2.read_text() == new_content_2

    sync_service_2.run_single_thread()
    sync_service_1.run_single_thread()

    assert file_path_1.read_text() == new_content_2
    assert file_path_2.read_text() == new_content_2


def test_delete_file(server_client: TestClient, datasite_1: SyftClientInterface, datasite_2: SyftClientInterface):
    server_settings: ServerSettings = server_client.app_state["server_settings"]
    sync_service_1 = SyncManager(datasite_1)
    sync_service_2 = SyncManager(datasite_2)

    # Setup initial state
    tree = {
        "folder1": {
            "_.syftperm": SyftPermission.mine_with_public_write(datasite_1.email),
            "file.txt": fake.text(max_nb_chars=1000),
        },
    }
    create_dir_tree(Path(datasite_1.datasite), tree)
    sync_service_1.run_single_thread()
    sync_service_2.run_single_thread()

    # delete the file
    file_path = datasite_1.datasite / "folder1" / "file.txt"
    file_path.unlink()

    sync_service_1.run_single_thread()

    # file is deleted on server
    assert (server_settings.snapshot_folder / datasite_1.email / "folder1" / "file.txt").exists() is False

    sync_service_2.run_single_thread()
    assert (datasite_2.datasite / datasite_1.email / "folder1" / "file.txt").exists() is False

    # Check if the metadata is gone
    remote_state_1 = sync_service_1.get_datasite_states()[0].get_remote_state()
    remote_paths = {metadata.path for metadata in remote_state_1}
    assert Path(datasite_1.email) / "folder1" / "file.txt" not in remote_paths


def test_invalid_sync_to_remote(server_client: TestClient, datasite_1: SyftClientInterface):
    sync_service_1 = SyncManager(datasite_1)
    sync_service_1.run_single_thread()

    # random bytes 1 byte too large
    too_large_content = os.urandom((MAX_FILE_SIZE_MB * 1024 * 1024) + 1)
    tree = {
        "valid": {
            "_.syftperm": SyftPermission.mine_with_public_write(datasite_1.email),
            "file.txt": "valid content",
        },
        "invalid_on_modify": {
            "_.syftperm": SyftPermission.mine_with_public_write(datasite_1.email),
            "file.txt": "valid content",
        },
        "invalid_on_create": {
            "_.syftperm": "invalid permission",
            "file.txt": too_large_content,
        },
    }

    create_dir_tree(Path(datasite_1.datasite), tree)
    sync_service_1.enqueue_datasite_changes(datasite=DatasiteState(datasite_1, email=datasite_1.email))

    queue = sync_service_1.queue
    consumer = sync_service_1.consumer

    items_to_sync = []
    while not queue.empty():
        items_to_sync.append(queue.get())
    assert len(items_to_sync) == 6  # 3 files + 3 permissions

    for item in items_to_sync:
        decision_tuple = consumer.get_decisions(item)
        abs_path = item.data.local_abs_path

        should_be_valid = item.data.path.parent.name in ["valid", "invalid_on_modify"]
        print(f"path: {abs_path}, should_be_valid: {should_be_valid}, parent: {item.data.path.parent}")

        is_valid = decision_tuple.remote_decision.is_valid(abs_path=abs_path, show_warnings=True)
        assert is_valid == should_be_valid, f"path: {abs_path}, is_valid: {is_valid}"

    sync_service_1.run_single_thread()

    # Modify invalid_on_modify to be invalid
    file_path = datasite_1.datasite / "invalid_on_modify" / "file.txt"
    file_path.write_bytes(too_large_content)
    permission_path = datasite_1.datasite / "invalid_on_modify" / "_.syftperm"
    permission_path.write_text("invalid permission")

    sync_service_1.enqueue_datasite_changes(datasite=DatasiteState(datasite_1, email=datasite_1.email))
    items_to_sync = []
    while not queue.empty():
        items_to_sync.append(queue.get())
    assert len(items_to_sync) == 4  # 2 invalid files + 2 invalid permissions

    for item in items_to_sync:
        decision_tuple = consumer.get_decisions(item)
        abs_path = item.data.local_abs_path

        is_valid = decision_tuple.remote_decision.is_valid(abs_path=abs_path, show_warnings=True)
        assert not is_valid, f"path: {abs_path}, is_valid: {is_valid}"


def test_sync_invalid_local_environment(datasite_1: SyftClientInterface):
    sync_service = SyncManager(datasite_1)
    sync_service.sync_interval = 0.1
    sync_folder = Path(datasite_1.workspace.datasites)

    # Create a file in datasite_1
    tree = {
        "folder1": {
            "_.syftperm": SyftPermission.mine_with_public_read(datasite_1.email),
            "file.txt": fake.text(max_nb_chars=1000),
        },
    }
    create_dir_tree(Path(datasite_1.datasite), tree)

    # Start syncing in separate thread
    sync_service.start()
    time.sleep(sync_service.sync_interval * 2)
    assert sync_service.is_alive()

    # Deleting the previous state file stops the sync
    shutil.rmtree(sync_folder.as_posix())

    max_wait_time = 5
    start_time = time.time()
    while sync_service.is_alive() and time.time() - start_time < max_wait_time:
        time.sleep(0.1)
    assert not sync_service.is_alive()

    # Restarting is not possible
    sync_service.start()
    start_time = time.time()
    while sync_service.is_alive() and time.time() - start_time < max_wait_time:
        time.sleep(0.1)
    assert not sync_service.is_alive()


def test_skip_symlink(server_client: TestClient, datasite_1: SyftClientInterface):
    sync_service = SyncManager(datasite_1)
    sync_service.run_single_thread()

    apps_dir = datasite_1.workspace.apps
    datasite_dir = datasite_1.datasite

    folder_to_symlink = apps_dir / "folder_to_symlink"
    file_to_symlink = apps_dir / "file_to_symlink.txt"

    folder_to_symlink.mkdir()
    file_to_symlink.write_text("content")

    # Nothing to sync, no writes to datasites
    states = sync_service.get_datasite_states()
    assert len(states) == 1
    assert states[0].is_in_sync()

    # Make symlinks in datasite
    symlink_folder = datasite_dir / "symlinked_folder"
    symlink_file = datasite_dir / "symlinked_file.txt"

    symlink_folder.symlink_to(folder_to_symlink)
    symlink_file.symlink_to(file_to_symlink)

    states = sync_service.get_datasite_states()
    assert len(states) == 1
    assert states[0].is_in_sync()

    # Check if symlinks are not synced
    sync_service.run_single_thread()
    snapshot_folder = server_client.app_state["server_settings"].snapshot_folder
    assert not (snapshot_folder / datasite_1.email / "symlinked_folder").exists()
    assert not (snapshot_folder / datasite_1.email / "symlinked_file.txt").exists()


def test_skip_hidden_paths(server_client: TestClient, datasite_1: SyftClientInterface):
    sync_service = SyncManager(datasite_1)
    sync_service.run_single_thread()

    hidden_folder = datasite_1.datasite / ".hidden_folder"
    hidden_nested_file = hidden_folder / "subfolder" / "file.txt"
    hidden_file = datasite_1.datasite / ".hidden_file.txt"

    hidden_folder.mkdir()
    hidden_nested_file.parent.mkdir(parents=True)
    hidden_file.write_text("content")

    states = sync_service.get_datasite_states()
    assert len(states) == 1
    assert states[0].is_in_sync()

    sync_service.run_single_thread()
    snapshot_folder = server_client.app_state["server_settings"].snapshot_folder
    assert not (snapshot_folder / datasite_1.email / ".hidden_folder").exists()
    assert not (snapshot_folder / datasite_1.email / ".hidden_file.txt").exists()


@pytest.mark.skip("This is for manual testing")
def test_n_datasites(tmp_path: Path, server_client: TestClient, datasite_1: SyftClientInterface):
    server_settings: ServerSettings = server_client.app_state["server_settings"]
    print(server_settings.snapshot_folder)

    n = 2
    files_per_datasite = 8

    logger.debug(f"Creating {n} datasites")

    emails = [f"datasite{i}@openmined.org" for i in range(n)]
    other_datasites: list[SyftClientInterface] = [setup_datasite(tmp_path, server_client, email) for email in emails]

    other_datasite_trees = [
        {
            "folder": {
                "_.syftperm": SyftPermission.mine_with_public_read(email),
            }
        }
        for email in emails
    ]

    for tree in other_datasite_trees:
        tree["folder"].update({f"file_{i}.txt": fake.text(max_nb_chars=10_000) for i in range(files_per_datasite - 1)})

    for datasite, tree in zip(other_datasites, other_datasite_trees):
        create_dir_tree(Path(datasite.datasite), tree)

    logger.debug("Syncing datasites")
    for datasite in other_datasites:
        sync_service = SyncManager(datasite)
        sync_service.run_single_thread()

    time.sleep(1)

    logger.debug("syncing main datasite")
    sync_service_1 = SyncManager(datasite_1)
    states = sync_service_1.get_datasite_states()
    assert len(states) == n + 1
    for state in states:
        out_of_sync_permissions, out_of_sync_files = state.get_out_of_sync_files()
        # assert out_of_sync_permissions
        # assert out_of_sync_files

    sync_service_1.run_single_thread()

    print(server_client.app_state["server_settings"].snapshot_folder)


def test_sync_health_check(datasite_1: SyftClientInterface):
    sync_service = SyncManager(datasite_1)
    sync_service.check_server_sync_status()

    sync_service.client.server_client.headers["Authorization"] = "Bearer invalid_token"
    with pytest.raises(FatalSyncError):
        sync_service.check_server_sync_status()



from pathlib import Path
from typing import Annotated, Optional

from typer import Context, Option, Typer

app = Typer(
    name="SyftBox Server",
    pretty_exceptions_enable=False,
    add_completion=False,
    context_settings={"help_option_names": ["-h", "--help"]},
)


# Define options separately to keep the function signature clean
# fmt: off
SERVER_PANEL = "Server Options"
SSL_PANEL = "SSL Options"

PORT_OPTS = Option(
    "-p", "--port",
    rich_help_panel=SERVER_PANEL,
    help="Local port for the SyftBox client",
)
WORKERS_OPTS = Option(
    "-w", "--workers",
    rich_help_panel=SERVER_PANEL,
    help="Number of worker processes",
)
VERBOSE_OPTS = Option(
    "-v", "--verbose",
    is_flag=True,
    rich_help_panel=SERVER_PANEL,
    help="Enable verbose mode",
)
RELOAD_OPTS = Option(
    "--reload", "--debug",
    rich_help_panel=SERVER_PANEL,
    help="Enable debug mode",
)
EMAIL_OPTS = Option(
    "-e", "--email",
    rich_help_panel=SERVER_PANEL,
    help="Email to ban/unban",
)
SSL_KEY_OPTS = Option(
    "--key", "--ssl-keyfile",
    exists=True, file_okay=True, readable=True,
    rich_help_panel=SSL_PANEL,
    help="Path to SSL key file",
)
SSL_CERT_OPTS = Option(
    "--cert", "--ssl-certfile",
    exists=True, file_okay=True, readable=True,
    rich_help_panel=SSL_PANEL,
    help="Path to SSL certificate file",
)
# fmt: on


@app.callback(invoke_without_command=True)
def server(
    ctx: Context,
    port: Annotated[int, PORT_OPTS] = 5001,
    workers: Annotated[int, WORKERS_OPTS] = 1,
    verbose: Annotated[bool, VERBOSE_OPTS] = False,
    ssl_key: Annotated[Optional[Path], SSL_KEY_OPTS] = None,
    ssl_cert: Annotated[Optional[Path], SSL_CERT_OPTS] = None,
):
    """Run the SyftBox server"""

    if ctx.invoked_subcommand is not None:
        return

    # lazy import to improve CLI startup performance
    import uvicorn

    from syftbox.server.server import app as fastapi_app

    uvicorn.run(
        app=fastapi_app,
        host="0.0.0.0",
        port=port,
        log_level="debug" if verbose else "info",
        workers=workers,
        ssl_keyfile=ssl_key,
        ssl_certfile=ssl_cert,
    )


def main():
    app()


if __name__ == "__main__":
    main()



from pathlib import Path

from rich.console import Console
from rich.tree import Tree


def display_file_tree(root_dir: Path):
    def add_dir(tree: Tree, path: Path):
        for child in path.iterdir():
            if child.is_dir():
                sub_tree = tree.add(f"📁 {child.name}")
                add_dir(sub_tree, child)
            elif child.name == "_.syftperm":
                tree.add(f"🛡️ {child.name}")
            else:
                tree.add(f"📄 {child.name}")

    console = Console()
    file_tree = Tree(f"📁 {root_dir.name}")
    add_dir(file_tree, root_dir)

    console.print(file_tree)



import contextlib

from fastapi import FastAPI, Request
from fastapi.middleware.cors import CORSMiddleware
from starlette.middleware.base import BaseHTTPMiddleware

from syftbox.client.base import SyftClientInterface
from syftbox.client.routers import app_router, datasite_router, index_router


class NoCacheMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request: Request, call_next):
        response = await call_next(request)
        response.headers["Cache-Control"] = "no-cache, no-store, must-revalidate"
        response.headers["Pragma"] = "no-cache"
        response.headers["Expires"] = "0"
        return response


@contextlib.asynccontextmanager
async def lifespan(app: FastAPI):
    yield


def create_api(client: SyftClientInterface) -> FastAPI:
    app = FastAPI(lifespan=lifespan)

    app.add_middleware(
        CORSMiddleware,
        allow_origins=[
            "http://localhost",
            "http://localhost:5001",
            "http://localhost:8080",
            "http://localhost:8081",
            "http://localhost:8083",
            "https://syftbox.openmined.org",
        ],  # Allow localhost on any port and specific domain
        allow_credentials=True,
        allow_methods=["*"],  # Allow all HTTP methods
        allow_headers=["*"],  # Allow all headers
    )

    app.state.client = client

    # Include routers
    app.include_router(index_router.router, tags=["index"])
    app.include_router(datasite_router.router, prefix="/datasites", tags=["datasites"])
    app.include_router(app_router.router, prefix="/apps", tags=["apps"])

    app.add_middleware(NoCacheMiddleware)

    return app



import asyncio

import pytest
from loguru import logger

from tests.e2e.conftest import Client, E2EContext, Server


def deployment_config():
    return {
        "e2e_name": "launch",
        "server": Server(port=5001),
        "clients": [
            Client(name="user1", port=8080, server_port=5001),
            Client(name="user2", port=8081, server_port=5001),
        ],
    }


@pytest.mark.asyncio
@pytest.mark.parametrize("e2e_context", [deployment_config()], indirect=True, ids=["deployment"])
async def test_e2e_launch(e2e_context: E2EContext):
    logger.info(f"Starting E2E '{e2e_context.e2e_name}'")
    e2e_context.reset_test_dir()
    await e2e_context.start_all()
    await asyncio.sleep(15)

    for client in e2e_context.clients:
        assert client.datasite_dir.exists()
        assert client.api_dir.exists()
        assert client.public_dir.exists()
        assert client.api_data_dir("").exists()
